!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#define SPEEDOFLIGHT 299792.458

! ******************* !
! COSMOLOGICAL TOOLS  !
! ******************* !

!*********!
! MODULES !
!*********!

module to_integrate
  use numbers
  integer(kind=singleI),parameter::nvar=5  ! number of equations in the ODE system
end module to_integrate


module integrations
  use numbers
  use to_integrate
  
  integer(kind=singleI),parameter::nbin=200
  real(kind=doubleR),parameter::log_amin=-3.,log_amax=0.2,dloga=(log_amax-log_amin)/dble(nbin)
  real(kind=doubleR),dimension(nbin)::scalef,cosmtime,grow1,grow2,DE_EoS,prdist,dprdist
  logical,save::firsttime=.true.
  real(kind=doubleR)::r2
  real(kind=doubleR),dimension(nbin):: rv,lamv,dldr ! ,lst,rst
  integer(kind=singleR)::ntabulated
  real(kind=doubleR),allocatable,dimension(:)::tabulated_a,tabulated_w
  logical::tabulated_flag=.false.
end module integrations


!*****************!
! INITIALIZATIONS !
!*****************!

subroutine compute_cosmological_quantities()

! Computes the following functions:
!   Scale factor, growth first and second order LPT, cosmic time, proper distance
! on a grid of values to be interpolated

  use numbers
  use pinocchio_module
  use integrations
  implicit none

  real(kind=doubleR),dimension(nvar)::y
  real(kind=doubleR)::x1,x2,eps,h1,res,dpr,Omegak,Omegak2,norm,p,k
  integer(kind=singleI)::nstep,i
  real(kind=doubleR), external :: integrand_eos,integr_dpr,power

  integer(kind=singleI),parameter::unit=120
  character(len=1)::string

! read the tabulated Equation of State of the dark energy

  if (trim(TabulatedEoSfile)/="no") then
     tabulated_flag=.true.

     if (proc_id==0) then
        open(unit,file=trim(TabulatedEoSfile),status="old")
        ntabulated=0
        do while (.true.)
           read(unit,'(a1)',end=99)string
           if (string=="#") cycle
           ntabulated=ntabulated+1
        enddo
99      close(unit)
        call MPI_BCAST(ntabulated, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
        allocate(tabulated_a(ntabulated))
        allocate(tabulated_w(ntabulated))
        open(unit,file=trim(TabulatedEoSfile),status="old")
        do i=1,ntabulated
           read(unit,*)tabulated_a(i),tabulated_w(i)
        enddo
        close(unit)
        call MPI_BCAST(tabulated_a, ntabulated, MPI_DOUBLE_PRECISION, 0, MPI_COMM_WORLD, ierr)
        call MPI_BCAST(tabulated_w, ntabulated, MPI_DOUBLE_PRECISION, 0, MPI_COMM_WORLD, ierr)
     else
        call MPI_BCAST(ntabulated, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
        allocate(tabulated_a(ntabulated))
        allocate(tabulated_w(ntabulated))
        call MPI_BCAST(tabulated_a, ntabulated, MPI_DOUBLE_PRECISION, 0, MPI_COMM_WORLD, ierr)
        call MPI_BCAST(tabulated_w, ntabulated, MPI_DOUBLE_PRECISION, 0, MPI_COMM_WORLD, ierr)
     endif

! these two parameters are given unrealistic values
     DEw0=10.
     DEw1=10.

  else
     tabulated_flag=.false.
  endif

! for the cosmological model
  Omegak=1-Omega0-OmegaLambda
  Omegak2=sqrt(abs(Omegak))

! first compute the integrand for the DE EoS
  do i=1,nbin

     if (i<nbin) then
        x2=10.**(log_amin + i*dloga)
     else
        x2=10.**log_amax
     endif

     call integrate(integrand_eos,x2,1.0,res)
     DE_EoS(i) = res
     scalef(i) = log10(x2)

  enddo

! ICs for the runge-kutta integration
  x1=1.d-6               ! initial value of scale factor a
  y(1)=1.d0              ! initial value of dD/da
  y(2)=x1                ! initial value of D(a)
  y(3)=2.0/3.0*x1**1.5   ! initial value of t(a)*Hubble0
  y(4)=-6./7.*x1         ! initial value of the dD2/da
  y(5)=-3./7.*x1**2      ! intial value of D2(a)
  eps=1.d-6              ! accuracy of integration
  h1=1.d-6               ! initial guess of time-step

! loop on bins for calling odeint
  do i=1,nbin

     if (i<nbin) then
        x2=10.**(log_amin + i*dloga)
     else
        x2=10.**log_amax
     endif

! calculation of cosmic time, 1st and 2nd-order growth functions by integrating the ODEs
     call odeint(y,x1,x2,eps,h1,nstep)

!     scalef(i)   = log10(x2)
     cosmtime(i) = log10(y(3))
     grow1(i)    = log10(y(2))
     grow2(i)    = log10(-y(5))

! calculation the proper distance (Mpc) for a generic cosmology (i.e. flat, open or closed) 
! and generic equation of state for the DE component 

     call integrate(integr_dpr,0.,(1./x2)-1,dpr)

     if(abs(Omegak) < 1.e-4) then 
        prdist(i)=(SPEEDOFLIGHT/(Hubble100*100.))*dpr
        
     else if(Omegak <= -1.e-4) then 
        prdist(i)=(SPEEDOFLIGHT/(Hubble100*100.))/Omegak2*sin(Omegak2*dpr)
        
     else if(Omegak >= 1.e-4) then  
        prdist(i)=(SPEEDOFLIGHT/(Hubble100*100.))/Omegak2*sinh(Omegak2*dpr)
        
     endif
!     prdist(i)= log10(prdist(i))

     x1=x2

  end do

! derivative of prdist
  do i=1,nbin
     if (i==1) then
        dprdist(1)=(prdist(2)-prdist(1))/(10.**(-scalef(2))-10.**(-scalef(1)))
     else if (i==nbin) then
        dprdist(nbin)=(prdist(nbin)-prdist(nbin-1))/(10.**(-scalef(nbin))-10.**(-scalef(nbin-1)))
     else
        dprdist(i)=(prdist(i+1)-prdist(i-1))/(10.**(-scalef(i+1))-10.**(-scalef(i-1)))
     endif
  enddo

  call interpola(scalef,grow1,nbin,0.0,norm)

  grow2 = grow2-2*norm  ! normalization of second-order growth rate
  grow1 = grow1-norm

  
! initialization of mass variance
  call Lambda_start()

! write out cosmological quantities
  if (proc_id==0) then
     open(10,file='pinocchio.'//trim(run)//'.cosmology.out',status='unknown')
     write(10,'("# Cosmological quantities used in PINOCCHIO (h=",f8.4,")")')Hubble100
     write(10,'("# TIME-DEPENDENT QUANTITIES")')
     write(10,'("# 1: scale factor")')
     write(10,'("# 2: cosmic time (Gyr)")')
     write(10,'("# 3: growth factor")')
     write(10,'("# 4: 2nd-order growth factor")')
     write(10,'("# 5: dark energy EOS")')
     write(10,'("# 6: proper distance (true Mpc)")')
     write(10,'("# 7: d/dz of proper distance (true Mpc)")')
     write(10,'("# SCALE-DEPENDENT QUANTITIES")')
     write(10,'("# 8: scale (true Mpc)")')
     write(10,'("# 9: mass variance")')
     write(10,'("#10: d Log sigma^2 / d Log R")')
     write(10,'("#11: k (true Mpc^-1)")')
     write(10,'("#12: P(k)")')
     write(10,'("#")')
     do i=1,nbin
        k=10.**(-4+dble(i)/dble(nbin)*6)
        p=power(k)
        write(10,'(99(1x,g14.6))')10.**scalef(i),10.**cosmtime(i)*9.7779/Hubble100,10.**grow1(i), &
             -10.**grow2(i),10.**(-3*scalef(i)) * exp(3.*DE_EoS(i)),prdist(i),dprdist(i), &
             10.**rv(i),10**lamv(i),10.**dldr(i),k,p
     enddo
     close(10)
  endif
  
  firsttime=.false.
  return
end subroutine compute_cosmological_quantities


! *********************** !
! GROWTH of PERTURBATIONS !
! *********************** !

function grow(z)

! Linear growing mode, normalized as grow=1/(1+z) at high redshift

! DIMENSIONLESS

  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: grow
  real(kind=doubleR),intent(in)::z

! local variables
  real(kind=doubleR):: aa,bb

  if (firsttime) then
     call abort_code("Function grow called too early!")
  endif

  aa=log10(1./(1.+z))
  call interpola(scalef,grow1,nbin,aa,bb)
  grow=10.**bb

  return
end function grow


function time(z)

! Cosmological time

! DIMENSION: Gyr

  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: time
  real(kind=doubleR),intent(in)::z

! local variables
  real(kind=doubleR):: aa,bb

  if (firsttime) then
     call abort_code("Function time called too early!")
  endif

  aa=log10(1./(1.+z))
  call interpola(scalef,cosmtime,nbin,aa,bb)
  time=10.**bb*9.7779/Hubble100

  return
end function time


function inverse_grow(b)

! Redshift corresponding to the growing mode b

! DIMENSIONLESS

  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: inverse_grow
  real(kind=doubleR),intent(in)::b

! local variables
  real(kind=doubleR)::loga,logb
      
  if (firsttime) then
     call abort_code("Function inverse_grow called too early!")
  endif

  logb=log10(b)
  call interpola(grow1,scalef,nbin,logb,loga)
  inverse_grow=10.**(-loga)-1.

  return
end function inverse_grow


function scale(t)

! Scale factor corresponding to a cosmic time

! DIMENSION: Gyr

  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: scale
  real(kind=doubleR),intent(in)::t

! local variables
  real(kind=doubleR):: aa,bb

  if (firsttime) then
     call abort_code("Function scale called too early!")
  endif

  aa=log10(t/9.7779*Hubble100)
  call interpola(cosmtime,scalef,nbin,aa,bb)
  scale=10.**bb

  return
end function scale


! integrand function for the proper distance
function integr_dpr(z)
  use pinocchio_module
  use integrations
  implicit none

  real(kind=doubleR) :: integr_dpr
  real(kind=doubleR),intent(in)::z  
  real(kind=doubleR),external::Hubble

  integr_dpr=(1./Hubble(z))*Hubble100*100.

end function integr_dpr



function proper_dist(z)  

! proper distance r(z) at the redshift z
! DIMENSION: Mpc (Htrue)

  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: proper_dist
  real(kind=doubleR),intent(in)::z   
  
  ! local variables
  real(kind=doubleR):: bb  !,zzz
  
  if (firsttime) then
     call abort_code("Function proper_dist(z) called too early!")
  endif 
 !z=(1./scalef)-1

  bb=log10(1./(1.+z))
 
  call interpola(scalef,prdist,nbin,bb,proper_dist)
! proper_dist=10.**zzz

return
end function proper_dist



function inverse_proper_dist(cc) 

! Redshift corresponding to the proper distance cc
  
  use pinocchio_module
  use integrations
  implicit none

  real(kind=doubleR) :: inverse_proper_dist
  real(kind=doubleR),intent(in)::cc
! integer(kind=singleI)::i
! local variables
  real(kind=doubleR)::logdd !,logcc
  

  if (firsttime) then
     call abort_code("Function inverse_proper_dist called too early!")
  endif
     
!  logcc=-log10(cc) 
  
  call interpola(-prdist,scalef,nbin,-cc,logdd)
  inverse_proper_dist=10.**(-logdd)-1.
  
  return
end function inverse_proper_dist


function dproper_dist_dz(z)

! d/dz of proper distance r(z) at the redshift z
! DIMENSION: Mpc (Htrue)

  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: dproper_dist_dz
  real(kind=doubleR),intent(in)::z

  ! local variables
  real(kind=doubleR):: bb

  if (firsttime) then
     call abort_code("Function proper_dist(z) called too early!")
  endif

  bb=log10(1./(1.+z))

  call interpola(scalef,dprdist,nbin,bb,dproper_dist_dz)

  return
end function dproper_dist_dz



! *********************** !
! MASS VIARIANCE (LAMBDA) !
! *********************** !

function aLambda_once(r)

! Linear variance on scale r, with smoothing as specified by FILTER;
! performs the integral each time
! r: Mpc

! DIMENSIONLESS

  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: aLambda_once
  real(kind=doubleR),intent(in)::r

  real(kind=doubleR):: out,limit
  real(kind=doubleR),external::Lambda_integrand

  r2=r
  if (filter.eq.0) then
     limit=log(5./r)
  else if (filter.eq.1) then
     limit=log(1./r)
  else
     limit=log(10.9041/r)
  endif

  call integrate(Lambda_integrand,-6.9d0,limit,out)
  aLambda_once=out

  return
end function aLambda_once


function Lambda_integrand(logk)

! Integrand for evaluation of Lambda

  use integrations
  implicit none
  real(kind=doubleR) :: Lambda_integrand
  real(kind=doubleR),intent(in)::logk
  real(kind=doubleR),external::power,window

  real(kind=doubleR):: k

  k=exp(logk)

  if (k.lt.1.d8) then         
     Lambda_integrand=power(k)*window(k*r2)**2*k**3
  else
     Lambda_integrand=0.
  endif

  return
end function Lambda_integrand


subroutine Lambda_start()

! Calculates the integrals for the mass variance

  use pinocchio_module
  use integrations
  implicit none

  real(kind=doubleR),external::aLambda_once

  real(kind=doubleR),parameter:: rmin=-6.d0, dr=0.05d0
  integer(kind=singleI)::i
  real(kind=doubleR)::r
  logical,save::done=.false.

!  if (done) return

  do i=1,nbin

     rv(i)=(rmin+(i-.5)*dr)
     r=10.**rv(i)
     lamv(i)=log10(max(aLambda_once(r),tiny(r)))

!     rst(nbin-i+1)=rv(i)
!     lst(nbin-i+1)=lamv(i)

  enddo

  do i=1,nbin
     if (i.eq.1) then
        dldr(i)=(lamv(i+1)-lamv(i))/(rv(i+1)-rv(i))
     else if (i.eq.nbin) then
        dldr(i)=(lamv(i)-lamv(i-1))/(rv(i)-rv(i-1))
     else
        dldr(i)=(lamv(i+1)-lamv(i-1))/(rv(i+1)-rv(i-1))
     endif
  enddo

  done=.true.
  return
end subroutine Lambda_start


function aLambda(r)

! Linear variance on scale r, with smoothing as specified by Lambda_start
! r: Mpc 

! DIMENSIONLESS

  use pinocchio_module
  use integrations

  implicit none
  real(kind=doubleR) :: aLambda
  real(kind=doubleR),intent(in)::r

! local variables
  real(kind=doubleR)::rlog
 
  if (firsttime) then
     call abort_code("Function aLambda called too early!")
  endif

! smallest radius is assumed to be 1 pc
  rlog=log10(max(r,1.d-6))
  call interpola(rv,lamv,nbin,rlog,aLambda)
  aLambda=10.**aLambda

  return
end function aLambda


function Radius(v)

! Radius corresponding to variance v, with smoothing as specified 
! by Lambda_start

! DIMENSION: Mpc

  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: Radius
  real(kind=doubleR),intent(in)::v
  
! local variables
  real(kind=doubleR)::vlog

  if (firsttime) then
     call abort_code("Function Radius called too early!")
  endif

  vlog=-log10(max(v,tiny(v)))
!  call interpola(lst,rst,nbin,vlog,Radius)
  call interpola(-lamv,rv,nbin,vlog,Radius)
  Radius=10.**Radius

  return
end function Radius



! ********************** !
! COSMOLOGICAL FUNCTIONS !
! ********************** !

function Omega(z)

! Cosmological mass density parameter as a function of redshift

! DIMENSIONLESS

  use pinocchio_module
  implicit none
  real(kind=doubleR) :: Omega
  real(kind=doubleR),intent(in)::z
  real(kind=doubleR),external :: Hubble

  Omega = Omega0 * (1+z)**(3) * (Hubble(z)/100./Hubble100)**(-2)

  return
end function Omega


function Hubble(z)

! Hubble parameter as a function of redshift

! DIMENSION: km/s/Mpc

  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: Hubble
  real(kind=doubleR),intent(in)::z
  real(kind=doubleR),external::Omega
  real(kind=doubleR)::Esq,OmegaK,eqds

  OmegaK = 1.-Omega0-OmegaLambda
  if (DEw0==-1. .and. DEw1==0.) then
     Esq = Omega0*(1.+z)**3+OmegaK*(1+z)**2+OmegaLambda
  else
     call interpola(scalef,DE_EoS,nbin,log10(1./(1.+z)),eqds)
     Esq = Omega0*(1.+z)**3+OmegaK*(1.+z)**2+OmegaLambda*(1.+z)**3 * exp(3.*eqds)
  endif
  Hubble=100.*Hubble100*sqrt(Esq)

  return
end function Hubble


function hubble_Gyr(z)

! Hubble parameter as a function of redshift

! DIMENSION: Gyr^-1

  use numbers
  implicit none
  real(kind=doubleR) :: hubble_Gyr
  real(kind=doubleR),intent(in)::z
  real(kind=doubleR),external::Hubble

  Hubble_gyr=Hubble(z)/977.79d0

  return
end function hubble_Gyr


function fomega(z)

! Peebles' f(Omega) function, dlogb/dloga

! DIMENSIONLESS

  use numbers
  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: fomega
  real(kind=doubleR),intent(in)::z
  real(kind=doubleR),external::grow

  real(kind=doubleR)::dz,db,zup,zdn

  dz=.02
  db=.5*(grow(z-dz)-grow(z+dz))/dz
  fomega=db*(1.+z)/grow(z)


  return
end function fomega


function aMass(v)

! Mass at linear variance V

! DIMENSION: M_sun

  use pinocchio_module
  implicit none
  real(kind=doubleR) :: aMass
  real(kind=doubleR),intent(in)::v
  
  real(kind=doubleR)::r
  real(kind=doubleR),external::Radius

  r=Radius(v)
  select case (filter)
  case (0)
     aMass=(2*acos(-1.))**1.5*r**3
  case (1)
     aMass=6.*acos(-1.)**2*r**3
  case (2)
     aMass=4.*acos(-1.)*r**3/3.
  case default
     call abort_code('Unknown smoothing in function AMASS')
  end select

  aMass=aMass*rho !2.778d11*Hubble100**2*Omega0

  return
end function aMass


function aM_star(z)

! Typical collapsing mass at redshift z

! DIMENSION: M_sun

  use pinocchio_module
  implicit none
  real(kind=doubleR) :: aM_star
  real(kind=doubleR),intent(in)::z
  real(kind=doubleR),external::aMass,grow

  aM_star=aMass((delta_c/grow(z))**2)
  
  return
end function aM_star



! ************** !
! POWER SPECTRUM !
! ************** !


function window(x)

! Window function

!     N.B:  x =: kr !!!

!     filter = 0: Gaussian smoothing
!     filter = 1: SKS smoothing
!     filter = 2: top-hat smoothing

! DIMENSIONLESS

  use pinocchio_module
  implicit none
  real(kind=doubleR) :: window
  real(kind=doubleR),intent(in)::x
  real(kind=doubleR),external::theta

  select case (filter)
     case (0)                       ! Gaussian
        window=exp(-x*x/2.)
     case (1)                       ! sharp k-space
        window=theta(1.-x)
     case (2)                       ! top-hat
        if (x.lt.3.e-3) then
           window=1.
        else if (x.lt.10.9041) then
           window=3.*(sin(x)/x**3-cos(x)/x**2)
        else
           window=0.
        endif
     case default
        call abort_code ('unknown filter in function window')
   end select

  return
end function window


function power(fk)

  use pinocchio_module
  implicit none

  real(kind=doubleR) :: power
  real(kind=doubleR),intent(in) :: fk 

#ifndef DO_NOT_USE_GENIC

! in this case the function is a wrapper for the GENIC function

  integer(kind=singleI) :: j
  real(kind=doubleR) :: power_from_C,kappa
#ifndef UNDERSCORE
  integer(kind=singleI),external :: power_wrap
  common /comm_wrap/power_from_C, kappa

  j=power_wrap(fk)

#else
  integer(kind=singleI),external :: power_wrap_
  common /comm_wrap_/power_from_C, kappa

  j=power_wrap_(fk)

#endif

  power=power_from_C

#else

! this is the old power(k) function of pinocchio 1-2

! PowerSpectrum = [-3,1]:  power-law power spectrum
! PowerSpectrum >= 10.: Power spectrum according to Efstathiou, Bond & White (1992)   
!          (Gamma=PowerSpectrum-10. if PowerSpectrum>10)
! PowerSpectrum <=-10.: Power spectrum according Bardeen et al. (1986)
!          (Gamma=-PowerSpectrum-10. if PowerSpectrum<-10)

! NEW PowerSpectrum=7 : wmap cmbfast result

! N.B.: fk IS SUPPOSED TO BE IN Mpc^-1


! local variables
  real(kind=doubleR)::pk,gamma,q
  real(kind=doubleR),external::transf_EH

  real(kind=doubleR)::power_from_C, kappa
#ifndef UNDERSCORE
  common /comm_wrap/power_from_C, kappa
#else
  common /comm_wrap_/power_from_C, kappa
#endif

  if (fk.eq.0.) then
     power=0.
     return
  endif

  if (PowerSpectrum<=1.1 .and. PowerSpectrum>=0.9) then
! Eisenstein & Hu fit, PowerSpectrum gives the primordial index
     kappa=fk
     pk = fk**PowerSpectrum*transf_EH()**2
  else if (PowerSpectrum.ge.-3..and.PowerSpectrum.le.1.) then
     if (fk.gt.5.d-4.and.fk.lt.5.d4) then
        pk=fk**PowerSpectrum
     else
        pk=0.
     endif
  else if (PowerSpectrum.ge.10.) then
     if (PowerSpectrum.eq.10.) then
        gamma=Omega0*Hubble100
     else
        gamma=PowerSpectrum-10.
     endif
     q=fk/Hubble100/gamma
     pk=q/(1.+(6.4*q+(3.*q)**1.5+(1.7*q)**2)**1.13)**(2./1.13)
  else if (PowerSpectrum.le.-10.) then
     if (PowerSpectrum.eq.-10.) then
        gamma=Omega0*Hubble100
     else
        gamma=-PowerSpectrum-10.d0
     endif
     q=fk/Hubble100/gamma
     pk=log(1.+2.34*q)**2*(1.+3.89*q+(16.1*q)*(16.1*q)+(5.64*q)**3+ &
          (6.71*q)**4)**(-.5)/q
  else
     call abort_code('Wrong PowerSpectrum in function POWER')

  endif
  power=pk*csp

#endif

  return
end function power


function transf_EH()
  use pinocchio_module
  implicit none

  real(kind=doubleR)::transf_EH

  real(kind=doubleR),parameter::Teta_27=1.0104
  real(kind=doubleR)::pk,gamma,q,fk
  real(kind=doubleR)::Omegac,Oh2,b1,b2,zd,Rd,zeq,Req,keq,s,ks,alc,bec,f,Tc,beb,bno,kst,ksi,Tb,Tr,y,alb,Ob2,OB
  real(kind=doubleR),external::T0

  real(kind=doubleR)::power_from_C, kappa
#ifndef UNDERSCORE
  common /comm_wrap/power_from_C, kappa
#else
  common /comm_wrap_/power_from_C, kappa
#endif

! Eisenstein & Hu fit, PowerSpectrum gives the primordial index

  fk   = kappa
  OB   = max(Omegab,1.e-6)
  Omegac = Omega0-OB
  Oh2 = Omega0*Hubble100**2
  Ob2 = OB*Hubble100**2
  b1  = 0.313*Oh2**(-0.419)*(1+0.607*Oh2**0.674)
  b2  = 0.238*Oh2**0.223
  zd  = 1291.*Oh2**0.251*(1.+b1*Ob2**b2)/(1.+0.659*Oh2**0.828)
  Rd  = 31.5*Ob2/(Teta_27**4*0.001*zd)
  zeq = 2.5e4*Oh2/Teta_27**4
  Req = 31.5*Ob2/(Teta_27**4*0.001*zeq)
  keq = 7.46d-2*Oh2/Teta_27**2
  s   = 1.633*log((sqrt(1.+Rd)+sqrt(Rd+Req))/(1+sqrt(Req)))/(keq*sqrt(Req))  ! 2sqrt(6)/3=1.633
  ks  = fk*s
  q   = fk*Teta_27**2/Oh2
  alc = ((46.9*Oh2)**0.670*(1.+(32.1*Oh2)**(-0.532)))**(-OB/Omega0) * &
       ((12.0*Oh2)**0.424*(1.+(45.0*Oh2)**(-0.582)))**(-(OB/Omega0)**3)
  bec = 1./(1.+(0.944/(1.+(458.*Oh2)**(-0.708)))*((Omegac/Omega0)**((0.395*Oh2)**(-0.0266))-1.))
  f   = 1./(1+(ks/5.4)**4)
  Tc  = f*T0(q,1.,bec)+(1.-f)*T0(q,alc,bec)
  beb = 0.5+OB/Omega0+(3.-2.*OB/Omega0)*sqrt((17.2*Oh2)**2+1.)
  bno = 8.41*Oh2**0.435
  kst = ks/(1.+(bno/ks)**3)**(0.3333)
  ksi = 1.6*Ob2**0.52*Oh2**0.73*(1.+(10.4*Oh2)**(-0.95))
  y   = (1.+zeq)/(1+zd)
  alb = 2.07*keq*s*(1+Rd)**(-0.75)*(y*(-6.*sqrt(1.+y)+(2.+3.*y)*log((sqrt(1.+y)+1.)/(sqrt(1.+y)-1.))))
  Tb  = (T0(q,1.,1.)/(1.+(ks/5.2)**2)+alb/(1.+(beb/ks)**3)*exp(-(fk/ksi)**1.4))*sin(kst)/kst
  Tr  = (OB*Tb+Omegac*Tc)/Omega0

  transf_EH=Tr

  return
end function transf_EH


function T0(q,a,b)
  use numbers
  implicit none
  real(kind=doubleR) :: T0
  real(kind=doubleR),intent(in)::q,a,b
  real(kind=doubleR)::ll,C

  ll = log(exp(1.)+1.8*b*q)
  C  = 14.2/a+386./(1.+69.9*q**1.08)
  T0 = ll/(ll+C*q**2)
  return
end function T0


subroutine init_cosmology()

  use pinocchio_module
  implicit none

  filter=2
  call normalization()

  rho = 2.775499745d11 * Hubble100**2 * Omega0

  filter=0
  call compute_cosmological_quantities()

  return
end subroutine init_cosmology


subroutine normalization()

! Normalization of the power spectrum to sigma_8

  use pinocchio_module
#ifndef DO_NOT_USE_GENIC
  use fmax_fftw_variables
  use fmax_grids
#endif
  implicit none

#ifndef DO_NOT_USE_GENIC
#ifndef UNDERSCORE
  integer(kind=singleI),external::genic
#else
  integer(kind=singleI),external::genic_
#endif
  integer(kind=singleI)::j
#else
  integer(kind=singleI)::old_filter
  real(kind=doubleR),external::aLambda_once
#endif

  csp = 1.d0

#ifdef DO_NOT_USE_GENIC
  old_filter = filter
  filter     = 2
  csp        = Sigma8*Sigma8/aLambda_once(8.d0/Hubble100)
  filter     = old_filter
#else

#ifndef UNDERSCORE
  j=GenIC(npart_fft, local_ny_after_transform, local_y_start_after_transform, proc_id, nproc, 1)
#else
  j=GenIC_(npart_fft, local_ny_after_transform, local_y_start_after_transform, proc_id, nproc, 1)
#endif

#endif

  return
end subroutine normalization



! ************* !
! MASS FUNCTION !
! ************* !


function Size(m)

! Radius corresponding to mass m
! m: M_sun

! DIMENSION: Mpc

  ! use run_module
  use pinocchio_module
  implicit none
  real(kind=doubleR) :: Size
  real(kind=doubleR),intent(in)::m

!  real(kind=doubleR)::rho
!  rho = 2.778d11*Hubble100**2*Omega0

  select case (filter)
  case (0)
     Size=(m/(2.d0*pi)**1.5d0/rho)**(1.d0/3.d0)
  case (1)
     Size=(m/(6.d0*pi**2*rho))**(1.d0/3.d0)
  case (2)
     Size=(m/(4.d0*pi*rho/3.d0))**(1.d0/3.d0)
  case default
     call abort_code('Unknown smoothing in function SIZE')
  end select

  return
end function Size


function Variance(m)

! Variance corresponding to mass m
! m: M_sun

! DIMENSIONLESS

  ! use run_module
  use pinocchio_module
  implicit none
  real(kind=doubleR) :: Variance
  real(kind=doubleR),intent(in)::m
  real(kind=doubleR),external::alambda

! local variables
  real(kind=doubleR)::r !,rho

!  rho = 2.778d11*Hubble100**2*Omega0
  select case (filter)
  case (0)
     r=(m/(2.d0*pi)**1.5d0/rho)**(1.d0/3.d0)
  case (1)
     r=(m/(6.d0*pi**2*rho))**(1.d0/3.d0)
  case (2)
     r=(m/(4.d0*pi*rho/3.d0))**(1.d0/3.d0)
  case default
     call abort_code('Unknown smoothing in function VARIANCE')
  end select

  Variance=aLambda(r)

  return
end function Variance


function dLambda(r)

! d Log Lambda/d Log r on scale r, with smoothing as specified by Lambda_start
! r: Mpc

! DIMENSIONLESS

  ! use run_module
  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR) :: dLambda
  real(kind=doubleR),intent(in)::r

! local variables
  real(kind=doubleR)::rlog

  if (firsttime) then
     call Lambda_start()
     firsttime=.false.
  endif

  rlog=log10(max(r,tiny(r)))
  call interpola(rv,dldr,nbin,rlog,dLambda)

  return
end function dLambda


function dodl(v,z)

! dOmega/dLambda

!     flag_ps=0:  Press  & Schechter (1974)
!     flag_ps=1:  Sheth & Tormen (2001)
!     flag_ps=2:  Jenkins et al. (2001)
!     flag_ps=3:  Warren et al. (2006)
!     flag_ps=4:  Reed et al. (2007)
!     flag_ps=5:  Crocce et al. (2010)
!     flag_ps=6:  Tinker et al. (2008)
!     flag_ps=7:  Courtin et al. (2010)

! DIFFERENTIALS: dLambda
! DIMENSIONLESS

  ! use run_module
  use pinocchio_module
  implicit none
  real(kind=doubleR) :: dodl
  real(kind=doubleR),intent(in)::v,z

! local variables
  real(kind=doubleR)::ni,sv,ni2,ni3,onepz
  real(kind=doubleR),parameter::sqrt2pi=0.39894228      ! c=1./sqrt(2.*pi)
  real(kind=doubleR),parameter::alpha=0.569558118758974
  sv=sqrt(v)
  ni=delta_c/sv
  
  select case (flag_ps)
  case (0)       ! Press & Schechter
     dodl=2.*exp(-.5d0*ni**2)*ni*sqrt2pi
  case (1)       ! Sheth & Tormen
     ni2=sqrt(0.707)*ni
     dodl=2.*0.3222d0*sqrt2pi*ni2*exp(-.5d0*ni2**2)*(1.d0+ 1.d0/ni2**0.6d0)
  case (2)       ! Jenkins et al.
     ni3=-log(sv)
     dodl=0.315d0*exp(-abs(ni3+0.61d0)**3.8d0)
  case (3)       ! Warren et al. (2006)
     dodl=0.7234d0*(sv**(-1.625d0)+0.2538d0)*exp(-1.1982d0/v)
  case (4)       ! Reed et al. (2007)
     ni2=sqrt(0.707)*ni
     dodl=2.*0.3222d0*sqrt2pi*ni2*exp(-.54d0*ni2**2)*(1.d0+ 1.d0/ni2**0.6d0 + 0.2d0*exp(-((-log(sv)-0.4d0)**2/0.72d0)))
  case (5)       ! Crocce et al. (2010)
     onepz=1.+min(1.,z)
     dodl=0.58*onepz**(-0.13)*(sv**(-1.37*onepz**(-0.15))+0.3*onepz**(-0.084))*exp(-1.036*onepz**(-0.024)/v)
  case (6)       ! Tinker et al. (2010)
     onepz=1.+min(2.5,z)
     dodl=0.186**onepz**(-0.14)*((2.57*onepz**(-alpha)/sv)**(1.47*onepz**(-0.06))+1.)*exp(-1.19/v)
  case (7)       ! Courtin et al. (2010)
     ni2=sqrt(0.695)*1.673/sv
     dodl=0.348*2.*sqrt2pi*ni2*(1.+(1./ni2**2)**0.1)*exp(-ni2**2/2.)
  case default
     call abort_code('Unknown flag_ps in function DODL')
  end select
         
  return
end function dodl


function an_h(mh,z)

! Number density of halos of mass mh (M_sun) at redshift z

! N.B.: LINEAR OUTPUT

! DIMENSION: Mpc^{-3} M_sun^{-1}
! DIFFERENTIALS: dMh

  ! use run_module
  use pinocchio_module
  implicit none
  real(kind=doubleR) :: an_h
  real(kind=doubleR),intent(in)::mh,z
  real(kind=doubleR),external::Size,Variance,grow,dodl,dLambda

! local variables
  real(kind=doubleR)::v,r

!  rho  = 2.778d11*Hubble100**2*Omega0
  v    = Variance(mh)*(grow(z))**2
  r    = Size(mh)
  an_h = rho*dodl(v,z)*abs(dLambda(r)/6.d0)/mh**2

  return
end function an_h



! *************** !
! OTHER FUNCTIONS !
! *************** !


function theta(x)

! Heavyside's theta function

  use numbers
  implicit none
  real(kind=doubleR) :: theta
  real(kind=doubleR),intent(in)::x

  if (x > 0.) then
     theta=1.d0
  else
     theta=0.d0
  endif
  return
end function theta


function asinh(x)

  use numbers
  implicit none
  real(kind=doubleR) :: asinh
  real(kind=doubleR),intent(in)::x

  asinh=log(x+sqrt(x*x+1.d0))

  return
end function asinh


function acosh(x)

  use numbers
  implicit none
  real(kind=doubleR) :: acosh
  real(kind=doubleR),intent(in)::x

  if (x.gt.1.) then
     acosh=log(x+sqrt(x*x-1.d0))
  else
!     print *,'Undefined ACOSH function'
     acosh=0.
  endif

  return
end function acosh


!Equation of state for DE (equation to integrate by qromb)
function integrand_eos(a)
  
  use pinocchio_module
  implicit none
  real(kind=doubleR),intent(in)::a
  real(kind=doubleR)::integrand_eos
  real(kind=doubleR),external::DE_EquationOfState

  integrand_eos = DE_EquationOfState(a)/a

  return
end function integrand_eos


function DE_EquationOfState(a)

  use pinocchio_module
  use integrations
  implicit none
  real(kind=doubleR)::DE_EquationOfState
  real(kind=doubleR),intent(in)::a

  if (tabulated_flag) then
     call interpola(tabulated_a,tabulated_w,ntabulated,a,DE_EquationOfState)
  else
     DE_EquationOfState=DEw0+(1-a)*DEw1
  endif

  return
end function DE_EquationOfState


! This subroutine gives the differentials of 1st- and 2nd-order growing mode 
! and of cosmic time
subroutine system_of_ODEs(x,y,dydx)

  use integrations
  use pinocchio_module
  implicit none

  REAL(kind=doubleR):: a1, b1, Esq, OmegaK!,eta,D2an,errore
  REAL(kind=doubleR)::  eqds, x, logx
  REAL(kind=doubleR), DIMENSION(nvar) :: y
  REAL(kind=doubleR), DIMENSION(nvar) :: dydx
  REAL(kind=doubleR), external :: integrand_eos, DE_EquationOfState

  OmegaK = 1.-Omega0-OmegaLambda

  if (DEw0==-1. .and. DEw1==0.) then
     
     Esq = Omega0/x**3+OmegaK/x**2+OmegaLambda
     a1   = -(3./x+1./2.*((-3.*Omega0/x**4-2*OmegaK/x**3)/Esq))
     b1   = 3./2.*Omega0/(Esq*x**5)

  else   ! Dark Energy is not a simple cosmological constant

     logx=log10(x)
     call interpola(scalef,DE_EoS,nbin,logx,eqds)

     Esq = Omega0/x**3+OmegaK/x**2+OmegaLambda/x**3 * exp(3.*eqds)
     a1   = -(3./x+1./2.*((-3.*Omega0/x**4-2*OmegaK/x**3-3*((1+DE_EquationOfState(x))/x**4) &
          *OmegaLambda*exp(3*eqds))/Esq))
     b1   = 3./2.*Omega0/(Esq*x**5)

  endif

  dydx(1) = a1*y(1)+b1*y(2)
  dydx(2) = y(1)
  dydx(3) = 1.0/x/sqrt(Esq)
  dydx(4) = a1*y(4)+b1*y(5)-b1*(y(2)**2)
  dydx(5) = y(4)
  
  return
end subroutine system_of_ODEs


SUBROUTINE odeint(y,x1,x2,eps,h1,nstep)
  
  use integrations
  use pinocchio_module
  IMPLICIT NONE

  integer(kind=singleI) :: nstepmax=1e5
  integer(kind=singleI),intent(out) :: nstep
  REAL(kind=doubleR), DIMENSION(nvar), INTENT(INOUT) :: y
  REAL(kind=doubleR), DIMENSION(nvar) :: dydx, y0
  REAL(kind=doubleR), INTENT(IN)      :: x1,eps
  REAL(kind=doubleR), INTENT(INOUT)   :: x2,h1
  real(kind=doubleR) :: x!,D2an,errore,eta,Esq,a,b,eqds,OmegaK
  real(kind=doubleR) :: h,hdid,hnext
  real,dimension(nvar)::yscal

  y0=y
  x=x1
  nstep=0
  h=min(h1,x2-x1)

  call system_of_ODEs(x,y,dydx)
 
  ciclo:  do while (x<x2)

     nstep=nstep+1

     yscal=abs(y)+abs(h*dydx)

!******************************************************************
     call rkqs(y,dydx,x,x2,h,eps,yscal,hdid,hnext)
!******************************************************************

     call system_of_ODEs(x,y,dydx)

     h=hnext

! Too many steps
     if (nstep>nstepmax) then
        write(*,'(''TOO MANY STEPS IN ODEINT! '',i9,9(1x,f12.9))') &
             nstepmax,x1,x,x2,h,h1
        exit ciclo
     endif

  enddo ciclo

  h1=h

  return
end SUBROUTINE odeint


SUBROUTINE rkqs(y,dydx,time,x2,htry,eps,yscal,hdid,hnext)

  use pinocchio_module
  use to_integrate  !contains the information of the number of equations to integrate
  IMPLICIT NONE
  REAL(kind=doubleR), DIMENSION(nvar), INTENT(INOUT) :: y
  REAL(kind=doubleR), DIMENSION(nvar), INTENT(IN)    :: dydx,yscal
  REAL(kind=doubleR), INTENT(INOUT) :: time
  REAL(kind=doubleR), INTENT(IN)    :: htry,eps,x2
  REAL(kind=doubleR), INTENT(OUT)   :: hdid,hnext

  REAL(kind=doubleR) :: errmax,h,htemp
  REAL(kind=doubleR) :: xnew
  REAL(kind=doubleR), DIMENSION(size(y)) :: yerr,ytemp
  REAL(kind=doubleR), PARAMETER :: SAFETY=0.9d0,PGROW=-0.2d0,PSHRNK=-0.25d0,&
       ERRCON=1.89d-4

  h=min(htry,x2-time)

  do while (.true.)

     call rkck(y,dydx,time,h,ytemp,yerr)

     errmax=maxval(abs(yerr(:)/yscal(:)))/eps

     if (errmax <= 1.0) exit
     htemp=SAFETY*h*(errmax**PSHRNK)

     h=sign(max(abs(htemp),0.1*abs(h)),h)
     h=min(h,(x2-time))
     xnew=time+h

     if (xnew == time) then
        if (proc_id==0) then
           write(*,*)'stepsize underflow in rkqs',time,xnew,htemp,h
           write(*,*)y,dydx
        endif
        call abort_code('stepsize underflow')
     endif

  end do

  if (errmax > ERRCON) then
     hnext=SAFETY*h*(errmax**PGROW)
  else
     hnext=5.0*h
  end if

  hdid=h
  time=time+h
  y(:)=ytemp(:)
  
  return
END SUBROUTINE rkqs

!Cash karp Runge Kutta step
SUBROUTINE rkck(y,dydx,x,h,yout,yerr)

  use to_integrate
  IMPLICIT NONE
  REAL(kind=doubleR), DIMENSION(nvar), INTENT(IN) :: y,dydx
  REAL(kind=doubleR), INTENT(IN) :: x
  REAL(kind=doubleR), INTENT(IN) :: h
  REAL(kind=doubleR), DIMENSION(nvar), INTENT(OUT) :: yout,yerr
  REAL(kind=doubleR), DIMENSION(size(y)) :: ak2,ak3,ak4,ak5,ak6,ytemp
  REAL(kind=doubleR), PARAMETER :: A2=0.2d0,A3=0.3d0,A4=0.6d0,A5=1.0d0, &
       A6=0.875d0,B21=0.2d0,B31=3.0d0/40.0d0,B32=9.0d0/40.0d0,&
       B41=0.3d0,B42=-0.9d0,B43=1.2d0,B51=-11.0d0/54.0d0,&
       B52=2.5d0,B53=-70.0d0/27.0d0,B54=35.0d0/27.0d0,&
       B61=1631.0d0/55296.0d0,B62=175.0d0/512.0d0,&
       B63=575.0d0/13824.0d0,B64=44275.0d0/110592.0d0,&
       B65=253.0d0/4096.0d0,C1=37.0d0/378.0d0,&
       C3=250.0d0/621.0d0,C4=125.0d0/594.0d0,&
       C6=512.0d0/1771.0d0,DC1=C1-2825.0d0/27648.0d0,&
       DC3=C3-18575.0d0/48384.0d0,DC4=C4-13525.0d0/55296.0d0,&
       DC5=-277.0d0/14336.0d0,DC6=C6-0.25d0

  ytemp=y+B21*h*dydx
  call system_of_ODEs(x+A2*h,ytemp,ak2)
  ytemp=y+h*(B31*dydx+B32*ak2) 
  call system_of_ODEs(x+A3*h,ytemp,ak3)
  ytemp=y+h*(B41*dydx+B42*ak2+B43*ak3)
  call system_of_ODEs(x+A4*h,ytemp,ak4)
  ytemp=y+h*(B51*dydx+B52*ak2+B53*ak3+B54*ak4)
  call system_of_ODEs(x+A5*h,ytemp,ak5)
  ytemp=y+h*(B61*dydx+B62*ak2+B63*ak3+B64*ak4+B65*ak5)
  call system_of_ODEs(x+A6*h,ytemp,ak6)
  yout=y+h*(C1*dydx+C3*ak3+C4*ak4+C6*ak6)
  yerr=h*(DC1*dydx+DC3*ak3+DC4*ak4+DC5*ak5+DC6*ak6)

  return
END SUBROUTINE rkck

