!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



subroutine allocate_vector()

! This subroutine allocates the vector necessary for Part 2 and 4 of the main code
! i.e. the computation of derivatives with fftw.

  use fmax_fftw_variables
  use pinocchio_module
  implicit none

! this check is hopefully useless; the allocate_vector should never be called
! if the vector is already allocated
  if (allocated(vector)) &
     call abort_code ('error in allocate_vector: vector already allocated')

! allocation of the fftw vector
  total_allocated = total_allocated + sizeof_vector*total_local_size
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating vector...')
  allocate(vector(total_local_size),stat=ierr)
  if (alloc_verbose) call message ('vector allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_VECTOR: unable to allocate vector')
  endif

! allocation of the fftw workspace
  total_allocated = total_allocated + sizeof_fftw_work*total_local_size
  max_allocated=max(total_allocated,max_allocated)
  if (allocated(fftw_work)) &
       call abort_code ('error in allocate_vector: fftw_work array already allocated')
  if (alloc_verbose) call message ('allocating fftw_work...')
  allocate(fftw_work(total_local_size),stat=ierr)
  if (alloc_verbose) call message ('fftw_work allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_VECTOR: unable to allocate workspace')
  endif

! message
  if(alloc_verbose) then
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  return
end subroutine allocate_vector


subroutine deallocate_vector()

! De-allocates the vectors allocated by allocate_vector

  use fmax_fftw_variables
  use pinocchio_module
  implicit none

  deallocate(vector)
  deallocate(fftw_work)

  total_allocated = total_allocated - sizeof_vector*total_local_size
  total_allocated = total_allocated - sizeof_fftw_work*total_local_size

  if(alloc_verbose) then
     write (info,"(' deallocate_vector, total allocated memory: ',f8.3,' Gbytes')") total_allocated/Gbyte
     call message(trim(info))
  end if

  return
end subroutine deallocate_vector


subroutine allocate_large_matrices()

! This subroutine allocates Fmax, Rmax, vmax and kdensity

  use fmax_fftw_variables
  use fmax_grids
  use pinocchio_module
  use final_products
  use fmax_main_memory
  implicit none

! allocation of Fmax
  total_allocated = total_allocated + sizeof_Fmax*npart_local
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating Fmax...')
  allocate(Fmax(0:Lgrid_local(1)-1,0:Lgrid_local(2)-1,0:Lgrid_local(3)-1),stat=ierr)
  if (alloc_verbose) call message ('Fmax allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_LARGE_MATRICES: unable to allocate vector')
  endif

! allocation of vmax
  total_allocated = total_allocated + sizeof_vmax*npart_local
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating vmax...')
  allocate(vmax(3,0:Lgrid_local(1)-1,0:Lgrid_local(2)-1,0:Lgrid_local(3)-1),stat=ierr)
  if (alloc_verbose) call message ('vmax allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_LARGE_MATRICES: unable to allocate vector')
  endif

! allocation of Rmax
  total_allocated = total_allocated + sizeof_Rmax*npart_local
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating Rmax...')
  allocate(Rmax(0:Lgrid_local(1)-1,0:Lgrid_local(2)-1,0:Lgrid_local(3)-1),stat=ierr)
  if (alloc_verbose) call message ('Rmax allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_LARGE_MATRICES: unable to allocate vector')
  endif

! allocation of kdensity
  total_allocated = total_allocated + sizeof_kdensity*npart_fft
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating kdensity...')
  allocate(kdensity(npart_fft),stat=ierr)
  if (alloc_verbose) call message ('kdensity allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_LARGE_MATRICES: unable to allocate vector')
  endif

!!!  if (ngrid>1) then
!!!     total_allocated = total_allocated + sizeof_kdensity*npart_fft(ngrid)
!!!     if (alloc_verbose) call message ('allocating kdensity_l...')
!!!     allocate(kdensity_l(npart_fft(ngrid)),stat=ierr)
!!!     if (alloc_verbose) call message ('kdensity_l allocated')
!!!     if (ierr>0) then
!!!        call abort_code('error in ALLOCATE_LARGE_MATRICES: unable to allocate vector')
!!!     endif
!!!  endif

! message
  if(alloc_verbose) then
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  return
end subroutine allocate_large_matrices


subroutine deallocate_some_large_matrices()

! This subroutine deallocates Rmax and kdensity

  use fmax_fftw_variables
  use fmax_grids
  use pinocchio_module
  use final_products
  use fmax_main_memory
  implicit none

! deallocation of Rmax
  total_allocated = total_allocated - sizeof_Rmax*npart_local
  deallocate(Rmax)
  if (alloc_verbose) call message ('Rmax deallocated')

! deallocation of kdensity
  total_allocated = total_allocated - sizeof_kdensity*npart_fft
  deallocate(kdensity)
  if (alloc_verbose) call message ('kdensity deallocated')

!!!  if (ngrid>1) then
!!!     total_allocated = total_allocated - sizeof_kdensity*npart_fft(ngrid)
!!!     deallocate(kdensity_l)
!!!     if (alloc_verbose) call message ('kdensity_l deallocated')
!!!  endif

! message
  if(alloc_verbose) then
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  return
end subroutine deallocate_some_large_matrices


subroutine deallocate_all_large_matrices()

! This subroutine deallocates Fmax and vmax

  use fmax_fftw_variables
  use fmax_grids
  use pinocchio_module
  use final_products
  use fmax_main_memory
  implicit none

! deallocation of Fmax
  total_allocated = total_allocated - sizeof_Fmax*npart_local
  deallocate(fmax)
  if (alloc_verbose) call message ('Fmax deallocated')

! deallocation of vmax
  total_allocated = total_allocated - sizeof_vmax*npart_local
  deallocate(vmax)
  if (alloc_verbose) call message ('vmax deallocated')

! message
  if(alloc_verbose) then
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  return
end subroutine deallocate_all_large_matrices


subroutine allocate_first_derivatives()
  use fmax_fftw_variables
  use fmax_grids
  use pinocchio_module
  use fmax_main_memory
  implicit none

! allocation of first_derivatives
  total_allocated = total_allocated + sizeof_firstd*npart_local
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating first derivatives...')
  allocate(first_derivatives(3,0:Lgrid_local(1)-1,0:Lgrid_local(2)-1,0:Lgrid_local(3)-1),stat=ierr)
  if (alloc_verbose) call message ('first derivatives allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_FIRST_DERIVATIVES: unable to allocate first derivatives')
  endif

!!!  if (ngrid>1) then
!!!! allocation of first_derivatives of large-scale box
!!!     total_allocated = total_allocated + sizeof_firstd*npart_local(ngrid)
!!!     if (alloc_verbose) call message ('allocating first derivatives for large box...')
!!!     allocate(first_derivatives_l(3,Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),Lgrid_local(ngrid,3)),stat=ierr)
!!!     if (alloc_verbose) call message ('vector allocated')
!!!     if (ierr>0) then
!!!        call abort_code('error in ALLOCATE_VECTOR: unable to allocate vector')
!!!     endif
!!!  endif

! message
  if(alloc_verbose) then
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  return
end subroutine allocate_first_derivatives


subroutine deallocate_first_derivatives
  use pinocchio_module
  use fmax_grids
  use fmax_main_memory
  implicit none

  deallocate(first_derivatives)
  total_allocated = total_allocated - sizeof_firstd*npart_local
!!!  if (allocated(first_derivatives_l)) then
!!!     deallocate(first_derivatives_l)
!!!     total_allocated = total_allocated - sizeof_firstd*npart_local(ngrid)
!!!  endif

  if(alloc_verbose) then
     write (info,'('' de-allocated first derivatives, total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  end if
  
  return
end subroutine deallocate_first_derivatives


subroutine allocate_second_derivatives

  use fmax_fftw_variables
  use fmax_grids
  use pinocchio_module
  use fmax_main_memory
  implicit none

! allocation of first_derivatives
  total_allocated = total_allocated + sizeof_secondd*npart_local
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating second_derivatives...')
  allocate(second_derivatives(6,0:Lgrid_local(1)-1,0:Lgrid_local(2)-1,0:Lgrid_local(3)-1),stat=ierr)
  if (alloc_verbose) call message ('second_derivatives allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_SECOND_DERIVATIVES: unable to allocate second derivatives')
  endif

!!!  if (ngrid>1) then
!!!! allocation of first_derivatives of large-scale box
!!!     total_allocated = total_allocated + sizeof_secondd*npart_local(ngrid)
!!!     if (alloc_verbose) call message ('allocating second_derivatives_l for large box...')
!!!     allocate(second_derivatives_l(6,Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),Lgrid_local(ngrid,3)),stat=ierr)
!!!     if (alloc_verbose) call message ('second_derivatives_l allocated')
!!!     if (ierr>0) then
!!!        call abort_code('error in ALLOCATE_VECTOR: unable to allocate vector')
!!!     endif
!!!  endif

! message
  if(alloc_verbose) then
     write (info,'('' SD, total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  return
end subroutine allocate_second_derivatives


subroutine deallocate_second_derivatives

  use fmax_main_memory
  use fmax_grids
  use pinocchio_module
  implicit none

  deallocate(second_derivatives)
  total_allocated = total_allocated - sizeof_secondd*npart_local
!!!  if (allocated(second_derivatives_l)) then
!!!     deallocate(second_derivatives_l)
!!!     total_allocated = total_allocated - sizeof_secondd*npart_local(ngrid)
!!!  endif

  if(alloc_verbose) then
     write (info,'('' de-allocated second derivatives, total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  end if
  
  return
end subroutine deallocate_second_derivatives


subroutine allocate_density
  use fmax_fftw_variables
  use fmax_grids
  use pinocchio_module
  use fmax_main_memory
  implicit none

! allocation of density
  total_allocated = total_allocated + sizeof_density*npart_local
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating density...')
  allocate(density(0:Lgrid_local(1)-1,0:Lgrid_local(2)-1,0:Lgrid_local(3)-1),stat=ierr)
  if (alloc_verbose) call message ('vector allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_DENSITY: unable to allocate density')
  endif

!!!  if (ngrid>1) then
!!!! allocation of density of large-scale box
!!!     total_allocated = total_allocated + density*npart_local(ngrid)
!!!     if (alloc_verbose) call message ('allocating density for large box...')
!!!     allocate(density_l(Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),Lgrid_local(ngrid,3)),stat=ierr)
!!!     if (alloc_verbose) call message ('vector allocated')
!!!     if (ierr>0) then
!!!        call abort_code('error in ALLOCATE_VECTOR: unable to allocate vector')
!!!     endif
!!!  endif

! message
  if(alloc_verbose) then
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  return
end subroutine allocate_density


subroutine deallocate_density

  use pinocchio_module
  use fmax_main_memory
  use fmax_grids
  implicit none

  deallocate(density)
  total_allocated = total_allocated - sizeof_density*npart_local
!!!  if (allocated(density_l)) then
!!!     deallocate(density_l)
!!!     total_allocated = total_allocated - sizeof_density*npart_local(ngrid)
!!!  endif

  if(alloc_verbose) then
     write (info,'('' de-allocated density, total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  end if
  
  return
end subroutine deallocate_density


subroutine allocate_smoothing_radii()

! Allocates vectors for smoothing radii

  use pinocchio_module
  use fmax_smoothing_radii
  implicit none

  if (alloc_verbose) call message ('allocating smoothing radii...')
  allocate(Smoothing_Radii(nsmooth))
  allocate(Smoothing_Variances(nsmooth))
  allocate(True_Variances(nsmooth)) ; True_Variances = 0.d0
  if (alloc_verbose) call message ('smoothing radii allocated')

  return
end subroutine allocate_smoothing_radii


subroutine allocate_VEL_for_displacements()
  use fmax_fftw_variables
  use fmax_grids
  use pinocchio_module
  use fmax_main_memory
  implicit none

! allocation of 1st-order term
  total_allocated = total_allocated + sizeof_firstd*npart_local
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating VEL_for_displacements...')
  allocate(VEL_for_displacements(3,0:Lgrid_local(1)-1,0:Lgrid_local(2)-1,0:Lgrid_local(3)-1),stat=ierr)
  if (alloc_verbose) call message ('VEL_for_displacements allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_VECTOR: unable to allocate VEL_for_displacements')
  endif

! message
  if(alloc_verbose) then
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  return
end subroutine allocate_VEL_for_displacements


subroutine deallocate_VEL_for_displacements
  use pinocchio_module
  use fmax_grids
  use fmax_main_memory
  implicit none

  deallocate(VEL_for_displacements)
  total_allocated = total_allocated - sizeof_firstd*npart_local
  if(alloc_verbose) then
     write (info,'('' de-allocated VEL_for_displacements, total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  end if
  
  return
end subroutine deallocate_VEL_for_displacements


subroutine allocate_fragment_vectors()

  use pinocchio_module
  use fragment_matrices
  implicit none

  total_allocated = total_allocated + sizeof_frag*Npart_frag
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating Fmax_frag...')
  allocate (Fmax_frag     (Npart_frag),stat=ierr)
  if (ierr>0) then
     call abort_code('error in INIT_FRAGMENT: unable to allocate Fmax_frag')
  endif
  if (alloc_verbose) call message ('Fmax_frag allocated')

  total_allocated = total_allocated + sizeof_frag*Npart_frag
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating vx...')
  allocate (vx       (Npart_frag),stat=ierr)
  if (ierr>0) then
     call abort_code('error in INIT_FRAGMENT: unable to allocate vx')
  endif
  if (alloc_verbose) call message ('vx allocated')

  total_allocated = total_allocated + sizeof_frag*Npart_frag
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating vy...')
  allocate (vy       (Npart_frag),stat=ierr)
  if (ierr>0) then
     call abort_code('error in INIT_FRAGMENT: unable to allocate vy')
  endif
  if (alloc_verbose) call message ('vy allocated')

  total_allocated = total_allocated + sizeof_frag*Npart_frag
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating vz...')
  allocate (vz       (Npart_frag),stat=ierr)
  if (ierr>0) then
     call abort_code('error in INIT_FRAGMENT: unable to allocate vz')
  endif
  if (alloc_verbose) call message ('vz allocated')

  return
end subroutine allocate_fragment_vectors


subroutine deallocate_fragment_vectors()

  use pinocchio_module
  use fragment_matrices
  implicit none

  total_allocated = total_allocated - sizeof_frag*Npart_frag
  deallocate (Fmax_frag)
  if (alloc_verbose) call message ('Fmax_frag deallocated')

  total_allocated = total_allocated - sizeof_frag*Npart_frag
  deallocate (vx)
  if (alloc_verbose) call message ('vx deallocated')

  total_allocated = total_allocated - sizeof_frag*Npart_frag
  deallocate (vy)
  if (alloc_verbose) call message ('vy deallocated')

  total_allocated = total_allocated - sizeof_frag*Npart_frag
  deallocate (vz)
  if (alloc_verbose) call message ('vz deallocated')

  return
end subroutine deallocate_fragment_vectors

