!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

! This file contains the routines that initialize the fmax run


subroutine initialization()

! Initializations for the Fmax run

  use pinocchio_module
  implicit none

! acquires the run basic information
  call set_parameters()

! checks that the filesystem is ready for the run
  call check_fs()

! sets the information for all the grids to be used and creates fftw plans
  call set_grids()

! sets the cosmological parameters and normalizes the power spectrum
  call init_cosmology()

! sets the properties of the PLC reconstruction (if compilation flag is set)
  call set_PLC()

! sets the smoothing radii
  call set_smoothing_radii()

! tests that there is enough space for allocations
  call test_allocations()

! allocates all the large matrices
  call allocate_large_matrices()

  !if (StartFromSavedFmax) return

! generates the initial density field
  call generate_densities()

! bye!
  return
end subroutine initialization


subroutine set_parameters()

! Reads all the basic information for the run
! file: fmax.info         contains: basic passwords for the run
! file: $run.info         contains: cosmology and box size
! file: $run.grids        contains: info on the other grids
! file: $run.pos          contains: positions to place the smallest box

  use pinocchio_module
  use fmax_grids
  use fmax_smoothing_radii

  implicit none

  character(len=300)::xParamfile,RunFlag,xOutputList !,xReadDensityField, xStartFromSavedFmax
  real(kind=doubleR)::xBoxSize,xOmega0,xOmegaLambda,xHubble100,xSigma8
  real(kind=doubleR)::xPowerSpectrum,xOmegaBaryon,xDEw0,xDEw1,xStartingzForPLC
  integer(kind=singleI)::xLgrid(3),xRandomSeed,xResamplingRate,xnout
  logical::xWriteFmax, xWriteVmax, xWriteRmax, xWriteDensity, xWriteDisplacements, &
       xCatalogInAscii, xDoNotFragment, xOutputInH100, BoxInH100
  real(kind=doubleR),dimension(1000)::zout_tmp
  real(kind=doubleR):: InputSpectrum_UnitLength_in_cm, WDM_PartMass_in_kev
  character(len=300)::FileWithInputSpectrum,xFileWithInputSpectrum
  character(len=300)::xTabulatedEoSfile

#ifndef UNDERSCORE
  common /params/ &
#else
  common /params_/ &
#endif
       xOmega0, xOmegaLambda, xHubble100, xSigma8,                 &
       xPowerSpectrum, xBoxSize, xOmegaBaryon, xDEw0, xDEw1, xStartingzForPLC,    &
       InputSpectrum_UnitLength_in_cm, WDM_PartMass_in_kev,  &
       xLgrid, xRandomSeed, xResamplingRate,           &
       xWriteFmax, xWriteVmax, xWriteRmax, xWriteDensity, xWriteDisplacements, &
       xCatalogInAscii, xDoNotFragment, xOutputInH100, BoxInH100, &
       xParamfile, RunFlag, xOutputList, &
       xFileWithInputSpectrum, xTabulatedEoSfile !, xReadDensityField, xStartFromSavedFmax


#ifndef UNDERSCORE
  common /outputs/ zout_tmp, xnout
#else
  common /outputs_/ zout_tmp, xnout
#endif

  integer(kind=singleI):: i,j,ibox(3)
#ifndef UNDERSCORE
  integer(kind=singleI),external::read_parameter_file
#else
  integer(kind=singleI),external::read_parameter_file_
#endif
  integer(kind=singleI),external::find_length,find_start

! read name of parameter file from command line
  call getarg(1,Paramfile)

  if (Paramfile=='') then
     call message('Please give parameter file name as an argument')
     call abort_code('')
  endif

! adds a space to let the C code know where the string ends
  Paramfile=trim(Paramfile)//' '
  xParamfile=Paramfile

! read the paramter file through the C routine
! that distributes data structure to all tasks
#ifndef UNDERSCORE
  j=read_parameter_file(proc_id)
#else
  j=read_parameter_file_(proc_id)
#endif

  if (j>0) then
     write(info,'("Error in reading parameter file ",a)')Paramfile
     call abort_code(info)
  endif

  Omega0        = xOmega0
  OmegaLambda   = xOmegaLambda
  Hubble100     = xHubble100
  Sigma8        = xSigma8
  PowerSpectrum = xPowerSpectrum
  Omegab        = xOmegaBaryon
  DEw0          = xDEw0
  DEw1          = xDEw1
  StartingzForPLC = xStartingzForPLC

  Lgrid=xLgrid
!  nbox=xnbox
  RandomSeed=xRandomSeed
  ResamplingRate=xResamplingRate

  WriteFmax=xWriteFmax
  WriteVmax=xWriteVmax
  WriteRmax=xWriteRmax
  WriteDensity=xWriteDensity
  WriteDisplacements=xWriteDisplacements
  !StartFlag=xStartFromSavedFmax
  !StartFromSavedFmax=.not.(trim(StartFlag) .eq. "no")
  CatalogInAscii=xCatalogInAscii
  DoNotFragment=xDoNotFragment
  OutputInH100=xOutputInH100

  if (DoNotFragment) then
     WriteFmax=.true.
     WriteVmax=.true.
  endif

  run                   = trim(RunFlag)
  OutputList            = trim(xOutputList)
  !ReadDensityField      = trim(xReadDensityField)
  Paramfile             = trim(xParamfile)
  FileWithInputSpectrum = trim(xFileWithInputSpectrum)
  TabulatedEoSfile    = trim(xTabulatedEoSfile)

  ngrid=1
  saved_seed=abs(RandomSeed)

  k_cutoff=1.d30

! sets box size and particle mass
  if (BoxInH100) then
     InterPartDist = xBoxSize/Hubble100/dble(Lgrid(1))
     safe          = int(SafetyBorder/Hubble100/InterPartDist)+1
  else
     InterPartDist = xBoxSize/dble(Lgrid(1))
     safe          = int(SafetyBorder/InterPartDist)+1
  endif
  BoxSize(:)    = InterPartDist*Lgrid(:)
  Mpart         = 2.775499745d11 * Hubble100**2 * Omega0 *InterPartDist**3
 
! computes the number of sub-boxes for fragmentation
  call compute_number_of_subboxes()
! compute grid length for fragmentation
  call find_subbox(ibox)
  if (ibox(1)==-1) then
     Lgrid_frag=0
  else
     do i=1,3
        Lgrid_frag(i)=find_length(Lgrid(i),nbox(i),ibox(i))+2*safe
        nstart(i)=find_start(Lgrid(i),nbox(i),ibox(i))-safe
     enddo
  endif
  Npart_frag = product(Lgrid_frag)

  delta_c=1.686
  flag_ps=3   ! Warren et al.

! number of outputs
  nout=xnout
  allocate (zout(nout))
  allocate (fout(nout))
  zout=zout_tmp(1:nout)
  call sort(nout,zout)

  if (proc_id==0) then
     write(*,'("PARAMETER VALUES")')
     write(*,'("Parameter file             ",a)')trim(Paramfile)
     write(*,'("Omega0                     ",f12.5)')Omega0              
     write(*,'("OmegaLambda                ",f12.5)')OmegaLambda        
     write(*,'("OmegaBaryon                ",f12.5)')Omegab
     if (trim(TabulatedEoSfile)/="no") then
        write(*,'("Dark Energy EoS will be read from file ",a)')trim(TabulatedEoSfile)
     else
        write(*,'("Dark Energy EoS parameters ",f8.5,1x,f8.5)')DEw0,DEw1
     endif
     write(*,'("Hubble100                   ",f12.5)')Hubble100          
     write(*,'("Sigma8                      ",f12.5)')Sigma8             
     write(*,'("PowerSpectrum               ",f12.5)')PowerSpectrum      
     write(*,'("BoxSize (true Mpc)          ",3(1x,f12.5))')BoxSize            
     write(*,'("BoxSize (Mpc/h)             ",3(1x,f12.5))')BoxSize*Hubble100
     write(*,'("GridSize                    ",3(1x,i5))')Lgrid              
     write(*,'("Particle Mass (true Msun)   ",g12.5)')Mpart
     write(*,'("Particle Mass (Msun/h)     " ,g12.5)')Mpart*Hubble100
     write(*,'("Inter-part dist (true Mpc)  ",g12.5)')InterPartDist
     write(*,'("Inter-part dist (Mpc/h)     ",g12.5)')InterPartDist*Hubble100
     write(*,'("Boundary layer (gridpoints) ",i3)')safe       
     write(*,'("RandomSeed                  ",i12)')RandomSeed        
     write(*,'("CatalogInAscii              ",l1)')CatalogInAscii         
     write(*,'("BoxInH100                   ",l1)')BoxInH100       
     write(*,'("OutputInH100                ",l1)')OutputInH100       
     write(*,'("WriteFmax                   ",l1)')WriteFmax        
     write(*,'("WriteVmax                   ",l1)')WriteVmax        
     write(*,'("WriteRmax                   ",l1)')WriteRmax        
     write(*,'("WriteDensity                ",l1)')WriteDensity     
     if (WriteDensity) write(*,'("ResamplingRate             ",l1)')ResamplingRate
     write(*,'("WriteDisplacements          ",l1)')WriteDisplacements
     !write(*,'("StartFromSavedFmax          ",l1)')StartFromSavedFmax
     !write(*,'("Flag of saved Fmax          ",a)')trim(StartFlag)
     write(*,'("DoNotFragment               ",l1)')DoNotFragment
     write(*,'("RunFlag                     ",a)')run
     write(*,'("OutputList                  ",a)')trim(OutputList)
     !write(*,'("ReadDensityField            ",a)')trim(ReadDensityField)
     write(*,'("nout                        ",i3)')nout
     write(*,'("zout                        ",99(1x,f6.3))')zout

     write(*,*)
     write(*,'("GENIC parameters:")')
     write(*,'("InputSpectrum_UnitLength_in_cm ",f12.5)')InputSpectrum_UnitLength_in_cm
     write(*,'("FileWithInputSpectrum          ",a)')trim(FileWithInputSpectrum)
     write(*,'("WDM_PartMass_in_kev            ",f12.5)')WDM_PartMass_in_kev

     write(*,*)
     write(*,'("Number of sub-boxes per dimension:",3(1x,i4))')nbox
     write(*,'("Core 0 will work on a grid:       ",3(1x,i4))')Lgrid_frag
     write(*,'("Number of particles for core 0:   ",i12)')     Npart_frag
     write(*,'("The resolved box will be:         ",3(1x,i4))')Lgrid_frag-2*safe
     write(*,'("procs used for fragmentation:     ",i5," over ",i5)') product(nbox),nproc
     write(*,*)
     select case (flag_ps)
     case (0)
        write(*,*)'Using Press & Schechter (1974) for the analytic mass function'
     case (1)
        write(*,*)'Using Sheth & Tormen (2001) for the analytic mass function'
     case (2)
        write(*,*)'Using Jenkins et al. (2001) for the analytic mass function'
     case (3)
        write(*,*)'Using Warren et al. (2006) for the analytic mass function'
     case (4)
        write(*,*)'Using Reed et al. (2007) for the analytic mass function'
     case (5)
        write(*,*)'Using Crocce et al. (2010) for the analytic mass function'
     case (6)
        write(*,*)'Using Tinker et al. (2008) for the analytic mass function'
     case (7)
        write(*,*)'Using Courtin et al. (2010) for the analytic mass function'
     end select
     write(*,*)
  endif

! if DoNotFragment is chosen, then create .info files for fragment-2.2 code

  if (DoNotFragment) then
     if (proc_id==0) then
        open(10,file=trim(run)//'.info',status='unknown')
        write(10,'(f12.6)')BoxSize(1)
        write(10,'(f12.6)')Omega0
        write(10,'(f12.6)')OmegaLambda
        write(10,'(f12.6)')PowerSpectrum
        write(10,'(f12.6)')Sigma8
        write(10,'(f12.6)')Hubble100
        write(10,'(i5,1x,i5,1x,i5)')Lgrid
        close(10)

        open(10,file='fragment.info',status='unknown')
        write(10,'(a)')trim(run)
        write(10,'(a)')trim(run)
        write(10,'(f12.6)')zout(nout)
        write(10,'(i4)')nout-1
        write(10,'(199(f12.6,1x))')zout(1:nout-1)
        write(10,'(i1)')0
        write(10,'(i1)')3
        write(10,'(i1)')1
        write(10,'(i1)')0
        close(10)

     endif
  endif

!  call message('.. set run done')

! bye!
 return
end subroutine set_parameters


#ifdef PLC
subroutine set_PLC

  use pinocchio_module
  use fragment_group_vectors, only: PLCone_centre
  implicit none

  integer(kind=singleI)::seed
  real(kind=doubleR),external::random,proper_dist,grow,an_h,dproper_dist_dz
  real(kind=doubleR)::z,dz,M,r,drdz,expected


  if (StartingzForPLC<0.) then
     number_of_rep=0
     Flimit_plc=-1.
     NPLC_max=0
     if (proc_id==0) then
        write(*,'("Negative value of StartingzForPLC, no Past Light Cone output will be given")')
        write(*,*)
     endif
     return
  endif

  seed=-RandomSeed
  PLCone_centre(1)=random(seed)*Lgrid(1)
  PLCone_centre(2)=random(seed)*Lgrid(2)
  PLCone_centre(3)=random(seed)*Lgrid(3)

  number_of_rep=int(proper_dist(StartingzForPLC)/BoxSize(1))+1
  Flimit_plc=1./grow(StartingzForPLC)

! counts the expected number of groups in the PLC
  filter=2
  call Lambda_start()

  expected=0.0
  z=0.
  dz=0.01
  do while (z<StartingzForPLC)
     M=MIN_MASS_FOR_CAT*Mpart
     r=proper_dist(z)
     drdz=dproper_dist_dz(z)
     do while (M<1.e16)
        expected=expected+an_h(M,z)*M*0.01*r**2*drdz*dz*4.*pi
        M=M*1.01
     enddo
     z=z+dz
  enddo

  NPLC_max=int(expected/product(BoxSize)*InterPartDist**3*1.5*dble(Npart_frag))+100

  if (proc_id==0) then
     write(*,'("Past Light Cone will be centred on point (grid units, before rotation):",3(1x,f12.4))')PLCone_centre
     write(*,'("The reconstruction will start at redshift ",f12.6)')StartingzForPLC
     write(*,'("The box will be replicated ",i2,"^3 times to construct the PLC")')2*number_of_rep+1
     write(*,'("Expected number of objects in the PLC: ",i10)')int(expected)
     write(*,'("Task 0 will use NPLC_max=",i10)')NPLC_max
     write(*,*)
  endif

  filter=0
  call Lambda_start()

  return
end subroutine set_PLC
#else
subroutine set_PLC
  use pinocchio_module
  implicit none

  if (proc_id==0) then
     write(*,'("PLC flag at compilation was not set, no Past Light Cone output will be given")')
     write(*,*)
  endif

end subroutine set_PLC
#endif

subroutine set_grids()

! computes relevant quantities for handling multiple grids
! including the plans for fftw

  use pinocchio_module
  use fmax_grids
  implicit none

! initializes fftw and local grids
  call init_fftw(1)
  if (ngrid>1) call init_fftw(2)

!  call message('.. set_grids done')

  return
end subroutine set_grids


subroutine set_smoothing_radii()

! Decides the smoothing radii for the run

! we chose smoothing scales such that the variance of the
! smoothed density field (the square of the mass variance) 
! ranges between sigma_min and sigma_max, in
! logarithmic steps of width step_var

! var_min is chosen so that, at the final redshift, delta=1.686 is 
! an n-sigma event, i.e. var_min = (1.686/nsigma)**2/grow(zout(nout))**2

! var_max is estimated as the variance of the Gaussian-smoothed field
! at 1/6th of the inter-particle distance

! the last variance value refers to the unsmoothed field 

  use fmax_smoothing_radii
  use pinocchio_module

  implicit none

! local variables
  integer(kind=singleI)::i
  real(kind=doubleR),parameter::nsigma=6.0
  real(kind=doubleR),parameter::var_min_z0=(1.686/nsigma)**2
  real(kind=doubleR),parameter::step_var=0.15
  real(kind=doubleR)::rmin,rmax,var_min,var_max
  real(kind=doubleR),external::aLambda,Radius,grow

  character(len=24):: fdate
  external fdate

  var_min = var_min_z0/(grow(zout(nout)))**2
  rmax    = Radius(var_min)
  rmin    = InterPartDist/6.
  var_max = aLambda(rmin)
  nsmooth = (log10(var_max)-log10(var_min))/step_var+2

  write(info,'("[",a24,"] Min variance: ",f12.6,", max variance: ",f12.6,", number of smoothing radii: ",i3)') &
       fdate(), var_min,var_max,nsmooth
  call message(trim(info))

  call allocate_smoothing_radii()

  do i=1,nsmooth-1
     Smoothing_Variances(i)=10.**(log10(var_min)+step_var*(i-1))
     Smoothing_Radii(i)=Radius(Smoothing_Variances(i))
  enddo
  Smoothing_Variances(nsmooth)=aLambda(InterPartDist/10.)
  Smoothing_Radii(nsmooth)=0.

!  call message('.. set_smoothing_radii: done ')

  return
end subroutine set_smoothing_radii


subroutine test_allocations()


! MEMORY REQUIREMENTS:            ZA
! (bytes per particle)
!
! allocate_large_matrices         44
! allocate_first_derivatives      24
! allocate_second_derivatives     48
! allocate_density                 8
! allocate_vector                 16
! allocate_fragment_vectors       32  * overhead
!

! allocates and deallocates matrices to test that memory is enough
  use pinocchio_module
  use fmax_grids
  implicit none
  integer(kind=singleI)::i

  call assign_fftw_variables(1)

  call allocate_large_matrices()        !  44

  call allocate_second_derivatives()    !  92

  call allocate_vector()                ! 108

  call deallocate_vector()              !  92

  call deallocate_second_derivatives()  !  44

  call deallocate_some_large_matrices() !  32

  call allocate_fragment_vectors()      ! +32 * overhead

  call deallocate_all_large_matrices()  ! 

  call deallocate_fragment_vectors()    !


! writes memory requirements for each processor
  call message ('Allocation test:')
  write(info,'("Expected total memory requirement: ",f10.3," Gb")') &
          bytes_per_particle*product(Lgrid*1.0)/Gbyte
  call message(info)
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  do i=0,nproc-1
     if (proc_id==i) then
        write(*,'("Processor ",i4," hosts ",i4," planes and needs ",f7.3," Gb (",f5.1," bytes per particle), I expected ",f7.3," Gb. Overhead: ",f6.3", memory check: ",f7.3," Gb")') &
             proc_id,Lgrid_local(3),max_allocated/Gbyte,max_allocated/product(Lgrid_local*1.0), &
             bytes_per_particle*product(Lgrid_local*1.0)/Gbyte, &
             product((Lgrid_frag*1.0/(Lgrid_frag-2.*safe))),total_allocated

     endif
     call MPI_BARRIER(MPI_COMM_WORLD,ierr)
  enddo



! Max allocations during fmax:  ZA -> 108 (->110)

! Expected allocation during fragment:   ZA ->  44*overhead + 80*groups


  return
end subroutine test_allocations


subroutine check_fs()

! Checks that the filesystem is ready for the run

  use numbers
  use pinocchio_module
  use fmax_grids
  implicit none

  if (proc_id==0) then

     if (WriteFmax .or. WriteVmax .or. WriteRmax .or. WriteDensity .or. WriteDisplacements) then ! .or. StartFromSavedFmax) then
        call check_dir('Data')
     endif

  endif

  return
end subroutine check_fs


subroutine check_dir(filename)

! checks the existence of a directory and aborts if it does not exist

  use numbers
  implicit none
  character(len=*),intent(in)::filename
#ifndef UNDERSCORE
  integer(kind=singleI),external::direxist
#else
  integer(kind=singleI),external::direxist_
#endif

  integer(kind=singleI)::i,j

  i=len_trim(trim(filename))

#ifndef UNDERSCORE
  j=direxist(filename,i)
#else
  j=direxist_(filename,i)
#endif

  if (j==1) then
     call abort_code('exiting...')
  endif

  return
end subroutine check_dir



!subroutine read_density_files()
!
!! Reads the ICs if it is required
!  use fmax_grids
!  implicit none
!
!! local variables
!  integer(kind=singleI)::igrid
!  
!  do igrid= 1,ngrid
!     if (read_IC(igrid)) then
!
!! allocates the vector
!        call assign_fftw_variables(igrid)
!        call allocate_vector()
!
!! Reads the density file
!        call read_one_density_field(igrid,InputDensityFile(igrid))
!
!! forward transform
!        call forward_transform()
!
!! saves the result to a tmp file
!        call save_kspace_vector(igrid)
!
!        call deallocate_vector()
!
!     endif
!  enddo
!
!  return
!end subroutine read_density_files


subroutine compute_number_of_subboxes()

  use pinocchio_module
  use fragment_grids
  use fmax_grids
  implicit none

  integer(kind=singleI)::i,j,k,np2,i1,j1,k1
  integer(kind=doubleI)::NN,NN1
  real(kind=doubleR)::Mpart_per_proc,last

! finds the optimal number of sub-boxes to use for the fragmentation

  last=1.e10

!  do np2=nproc,nproc/2,-1

  np2=nproc

     do k=1,nproc  !/2
        do j=k,nproc   !/2
           do i=j,nproc    !/2

              if (i*j*k==np2) then

                 NN=int(i*j*k,8)*int(Lgrid(3)/k+1+safe,8)*int(Lgrid(2)/j+1+safe,8)*int(Lgrid(1)/i+1+safe,8)
                 Mpart_per_proc=dble(NN)/dble(i*j*k)/1024./1024.
                 if (Mpart_per_proc<last) then
                    last=Mpart_per_proc
                    NN1=NN
                    i1=i
                    j1=j
                    k1=k
                 endif
              end if

           end do
        end do
     end do

!  enddo

  nbox(1)=i1
  nbox(2)=j1
  nbox(3)=k1

  do i=1,3
     pbc_flag(i)=(nbox(i)==1)
  enddo

  if (product(nbox)/=nproc) then
     write(info,'("The optimal total number of sub-boxes (",i3,",",i3,",",i3,") is less than the number of processors")')nbox
     call message(info)
     call message('We are not ready to run in these conditions')
     call finish('Try to run Pinocchio again with a different number of processors')
  endif

  return
end subroutine compute_number_of_subboxes

