!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#define NYQUIST 1.0
!#define TZA


subroutine assign_fftw_variables(igrid)

! Assigns the variables to be passed to the fftw wrapper

  use fmax_fftw_variables
  use fmax_grids
  use pinocchio_module
  implicit none
  integer(kind=singleI),intent(in)::igrid

  if (igrid==1) then

     ThisGrid                      = 1
     nx_fftw                       = Lgrid(1)
     ny_fftw                       = Lgrid(2)
     nz_fftw                       = Lgrid(3)
     total_local_size              = npart_fft
     reverse_plan                  = saved_reverse_plan
     forward_plan                  = saved_forward_plan
     local_nz                      = Lgrid_local(3)
     local_z_start                 = Lgrid_z_start
     local_ny_after_transform      = Lgrid_fft(2)
     local_y_start_after_transform = Lgrid_y_start
     cell_size                     = InterPartDist
     lower_k_cutoff = 0.d0
!     upper_k_cutoff = 0.8 * pi
     upper_k_cutoff = NYQUIST * pi

  else

     call abort_code("MULTIPLE GRIDS NOT IMPLEMENTED YET")

  endif

  return
end subroutine assign_fftw_variables


subroutine compute_density(ismooth,igrid)

! Computes the density

  use pinocchio_module
  use fmax_grids
  use fmax_smoothing_radii
  use fmax_fftw_variables
  use fmax_main_memory
  implicit none
  integer(kind=singleI),intent(in)::ismooth,igrid

  integer(kind=singleI)::local_x,local_y,local_z
  integer(kind=doubleI)::index
  integer(kind=doubleI), external :: index_rspace

! allocates the vector for fft
  call assign_fftw_variables(igrid)
  call allocate_vector()

! the density is the laplacian of the potential
  if (igrid==1) then
     Rsm_grid=Smoothing_radii(ismooth)/InterPartDist
     vector=kdensity
  else
     Rsm_grid=0.0
     vector=kdensity_l
  endif

  call compute_derivative(-1,-1)  ! (-1,-1) sets the greens function to be 1

  if (igrid==1) then
     do local_z=0,local_nz-1
        do local_y=0, ny_fftw-1
           do local_x=0, nx_fftw-1
              index        = index_rspace(local_x, local_y, local_z)
              density(local_x,local_y,local_z) = vector(index)
           enddo
        enddo
     enddo
  else
     do local_z=0,local_nz-1
        do local_y=0, ny_fftw-1
           do local_x=0, nx_fftw-1
              index        = index_rspace(local_x, local_y, local_z)
              density_l(local_x,local_y,local_z) = vector(index)
           enddo
        enddo
     enddo
  endif

  call deallocate_vector()

  return
end subroutine compute_density


subroutine generate_densities()

! Computes the density

  use pinocchio_module
  use fmax_grids
  use fmax_smoothing_radii
  use fmax_fftw_variables
  use fmax_main_memory
  implicit none
  integer(kind=singleI)::igrid
  character(len=24):: fdate
  external fdate

  cputmp = MPI_WTIME()
  write(info,'(''['',a24,''] Generating density in Fourier space'')') fdate()
  call message (trim(info))

  do igrid=1,ngrid
! allocates the vector for fft

     call assign_fftw_variables(igrid)
     call allocate_vector()

! the density is the laplacian of the potential
     call generate_one_density()

     if (igrid==1) then
        kdensity=vector
     else
        kdensity_l=vector
     endif

     call deallocate_vector()

  enddo

  cputmp = MPI_WTIME() - cputmp    ! timing
  write(info,'(''['',a24,''] Done generating densities, cputime = '',g14.6)') fdate(),cputmp
  call message (trim(info))
  cputime(4)=cputime(4)+cputmp

  return
end subroutine generate_densities


subroutine compute_first_derivatives(Rsmooth,igrid)

! Computes the three components of the first derivative tensor
! of the peculiar rescaled gravitational potential;

! igrid= 1: small-scale box
! igrid= 2: large-scale box

  use pinocchio_module
  use fmax_grids
  use fmax_fftw_variables
  use fmax_main_memory
  use fmax_smoothing_radii

  implicit none
  integer(kind=singleI),intent(in)::igrid
  real(kind=doubleR),intent(in)::Rsmooth
!
  character(len=24):: fdate
  external fdate

! local variables
  integer(kind=singleI)::ia

  integer(kind=singleI) :: local_x, local_y, local_z
  integer(kind=doubleI) :: index
  integer(kind=doubleI), external :: index_rspace


! allocates the vector for fft
  if (igrid>0) then
     call assign_fftw_variables(igrid)
  else 
     call assign_fftw_variables(1)
  endif
  call allocate_vector()

! computes all three first derivatives of the potential
! and saves them (the saving strategy is decided in optimized.F90)
  do ia=1,3

     write(info,'(''['',a24,''] Compute 1st derivative : '',i4)') fdate(),ia
     call message (trim(info))

! only grid 1 must be smoothed
     Rsm_grid=0.0
     if (igrid==1) then
        vector=kdensity
        Rsm_grid=Rsmooth/InterPartDist
     else if (igrid==2) then
        vector=kdensity_l
     else
        call abort_code("WRONG igrid VALUE IN COMPUTE_FIRST_DERIVATIVES")
     endif

     call compute_derivative(ia,0)

!     write(info,'(''['',a24,''] Save derivative : '',i4)') fdate(),ia
!     call message (trim(info))

     do local_z=0,local_nz-1
        do local_y=0, ny_fftw-1
           do local_x=0, nx_fftw-1
              index        = index_rspace(local_x, local_y, local_z)
              first_derivatives(ia,local_x,local_y,local_z) = vector(index)
           enddo
        enddo
     enddo

  enddo

  call deallocate_vector()

  return
end subroutine compute_first_derivatives


subroutine compute_second_derivatives(Rsmooth,igrid)

! Computes the six components of the second derivative tensor
! of the peculiar rescaled gravitational potential;

  use pinocchio_module
  use fmax_grids
  use fmax_fftw_variables
  use fmax_main_memory
  use fmax_smoothing_radii

  implicit none
  integer(kind=singleI),intent(in)::igrid
  real(kind=doubleR),intent(in)::Rsmooth
!
  character(len=24):: fdate
  external fdate

! local variables
  integer(kind=singleI)::ia,ib,ider

  integer(kind=singleI) :: local_x, local_y, local_z
  integer(kind=doubleI) :: index
  integer(kind=doubleI), external :: index_rspace

! allocates the vector for fft
  if (igrid>0) then
     call assign_fftw_variables(igrid)
  else 
     call assign_fftw_variables(1)
  endif
  call allocate_vector()

! computes all six second derivatives of the potential
! and saves them (the saving strategy is decided in optimized.F90)

  do ia=1,3
     do ib=ia,3
        if (ia==ib) then
           ider=ia
        else
           ider=ia+ib+1
        endif

        write(info,'(''['',a24,''] Compute 2nd derivative : '',i4)') fdate(),ider
        call message (trim(info))

! only grid 1 must be smoothed
        Rsm_grid=0.0
        if (igrid==1) then
           vector=kdensity
           Rsm_grid=Rsmooth/InterPartDist
        else if (igrid==2) then
           vector=kdensity_l
        else
           call abort_code("WRONG igrid VALUE IN COMPUTE_SECOND_DERIVATIVES")
        endif


        call compute_derivative(ia,ib)

!        write(info,'(''['',a24,''] Save derivative : '',i4)') fdate(),ider
!        call message (trim(info))

        do local_z=0,local_nz-1
           do local_y=0, ny_fftw-1
              do local_x=0, nx_fftw-1
                 index        = index_rspace(local_x, local_y, local_z)
                 second_derivatives(ider,local_x,local_y,local_z) = vector(index)
              enddo
           enddo
        enddo

     enddo
  enddo
  call deallocate_vector()

  return
end subroutine compute_second_derivatives


subroutine compute_collapse_times(ismooth)

! Computes the collapse times for one smoothing radius and position
! by collecting information from all grids
! and then computing the collapse times

  use pinocchio_module
  use fmax_grids
  use fmax_smoothing_radii
  use final_products
  use fmax_main_memory

  implicit none
  integer(kind=singleI),intent(in)::ismooth

  real(kind=doubleR),external::inverse_collapse_time

! local variables
  integer(kind=singleI)::local_z,global_x,global_y,global_z,fail
  real(kind=doubleR):: Fnew
  real(kind=doubleR):: local_variance,global_variance,local_average,global_average,delta
  real(kind=doubleR)::diff_ten(6)

! initialize the variance of the slice
  local_variance = 0.d0
  local_average  = 0.d0

! initialize Fmax if it is the first smoothing
  if (ismooth == 1) then
     Fmax = -10.d0
     Rmax = -1
  endif

! loop on all particles in the slice
  do local_z=0,Lgrid_local(3)-1
     global_z=local_z+Lgrid_z_start
     do global_y=0,Lgrid_local(2)-1
        do global_x=0,Lgrid_local(1)-1

! NB: QUI BISOGNA SOMMARE SULLE DUE BOX

! sum all contributions
           diff_ten(1:6) = second_derivatives(:,global_x,global_y,local_z)

! Computation of the variance of the linear density field
           delta = sum(diff_ten(1:3))
           local_average  = local_average  + delta
           local_variance = local_variance + delta**2

! Computation of ellipsoidal collapse
           Fnew = inverse_collapse_time(diff_ten, fail)

           if (fail==1) then
              print *,proc_id,'Warning: failure in inverse_collapse_time',global_x,global_y,global_z
              Fnew=99.
           endif

           if (Fmax(global_x,global_y,local_z) < Fnew) then
              Fmax(global_x,global_y,local_z) = Fnew
              Rmax(global_x,global_y,local_z) = ismooth
           endif

        enddo
     enddo
  enddo

! Saves the obtained variance
  global_variance = 0.d0
  global_average  = 0.d0

  CALL MPI_REDUCE(local_variance, global_variance, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, MPI_COMM_WORLD, ierr)
  CALL MPI_REDUCE(local_average , global_average , 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, MPI_COMM_WORLD, ierr)

  if (proc_id == 0) then

     global_variance = global_variance / Lgrid(1) / Lgrid(2) / Lgrid(3)
     global_average  = global_average  / Lgrid(1) / Lgrid(2) / Lgrid(3)
     True_variances(ismooth) = global_variance
  endif

  call MPI_BCAST(global_variance, 1, MPI_DOUBLE_PRECISION, 0, MPI_COMM_WORLD, ierr)

  True_Variances(ismooth)=global_variance

  if (proc_id == 0) then
     write(*,'(" *** Smoothing radius: ",f15.6," - theorical variance: ",f15.6," - true variance:",f15.6)') &
          Smoothing_Radii(ismooth), Smoothing_Variances(ismooth), True_Variances(ismooth)
     if (DoNotFragment) then
        open(10,file=trim(run)//'.var',status='unknown', position='append')
        write(10,'(3(1x,f15.8))') Smoothing_Radii(ismooth) &
             , Smoothing_Variances(ismooth) &
             , True_Variances(ismooth)
        close(10)
     endif
  endif

! bye!
  return
end subroutine compute_collapse_times


function inverse_collapse_time(deformation_tensor, fail)

! Computes the inverse collapse time of a grid point
! given its deformation tensor
  use pinocchio_module

  implicit none
  real(kind=doubleR) :: inverse_collapse_time
  real(kind=doubleR),intent(in)::deformation_tensor(6)
  integer(kind=singleI),intent(out)::fail

! local variables
  integer(kind=singleI)::m
  real(kind=doubleR)::mu1,mu2,mu3,c,q,r,sq,t,x1,x2,x3
  real(kind=doubleR),external::ell

! mu1, mu2 and mu3 are the principal invariants of the 3x3 tensor 
! of second derivatives.


! diagonalization of the tensor
  mu1=sum(deformation_tensor(1:3))
  mu2=.5*mu1*mu1
  do m=1,6
     if (m.le.3) then 
        c=.5
     else
        c=1.
     endif
     mu2=mu2-c*deformation_tensor(m)*deformation_tensor(m)
  enddo
  mu3=deformation_tensor(1)*deformation_tensor(2)*deformation_tensor(3) &
    + 2.*deformation_tensor(4)*deformation_tensor(5)*deformation_tensor(6) &
       - deformation_tensor(1)*deformation_tensor(6)*deformation_tensor(6) &
       - deformation_tensor(2)*deformation_tensor(5)*deformation_tensor(5) &
       - deformation_tensor(3)*deformation_tensor(4)*deformation_tensor(4)

  q=(mu1*mu1-3.*mu2)/9.

  if (q==0.) then
! in this case the tensor is already diagonal
     x1=deformation_tensor(1)
     x2=deformation_tensor(2)
     x3=deformation_tensor(3)
  else
     r=-(2.*mu1**3-9.*mu1*mu2+27.*mu3)/54.

     if (q*q*q.lt.r*r.or.q.lt.0.) then
        fail=1
        inverse_collapse_time=-10.d0
        return
     endif

     if(q < 0)then
        write (*,*) ' error: q<0'
        call abort_code(' q < 0 in inverse_collapse_times')
     endif
     sq=sqrt(q)     
     t=acos(r/q/sq)
     x1=-2.*sq*cos(t/3.)+mu1/3.
     x2=-2.*sq*cos((t+2.*pi)/3.)+mu1/3.
     x3=-2.*sq*cos((t+4.*pi)/3.)+mu1/3.

  endif

  call ord(x1,x2,x3)
  t=ell(x1,x2,x3,mu1,mu3)
     
! inverse collapse time
     
  if (t>0.) then
     inverse_collapse_time = 1./t
  else
     inverse_collapse_time = -10.
  endif

  fail=0
  return
end function inverse_collapse_time


function ell(l1,l2,l3,del,det)
  use pinocchio_module

  implicit none
  real(kind=doubleR) :: ell
  real(kind=doubleR) :: l1,l2,l3,del,det

! This rouine computes the smallest non-negative solution of the 3rd
! order equation for the ellipsoid, and corrects it to reproduce the
! spherical collapse correctly.

! local variables
  real(kind=doubleR) :: a1,a2,a3,den,dis,r,q,sq,t,s1,s2,s3,corr

  den=det/126.+5.*l1*del*(del-l1)/84.
  if (l1.eq.0.) then
     ell=-.1
  else if (den.eq.0.) then
     if (del.eq.l1) then
        if (l1.gt.0) then
           ell=1./l1
        else
           ell=-.1
        endif
     else
        dis=7.*l1*(l1+6.*del)
        if (dis.lt.0.) then
           ell=-.1
        else
           ell=(7.*l1-sqrt(dis))/(3.*l1*(l1-del))
           if (ell.lt.0.) ell=-.1
        endif
     endif
  else
     a1=3.*l1*(del-l1)/14./den
     a2=l1/den
     a3=-1./den
     q=(a1*a1-3.*a2)/9.
     r=(2.*a1*a1*a1-9.*a1*a2+27.*a3)/54.
     if (r*r.gt.q*q*q) then
        sq=(sqrt(r*r-q*q*q)+abs(r))**(1./3.)
        ell=-abs(r)/r*(sq+q/sq)-a1/3.
        if (ell.lt.0.) ell=-.1
     else
        sq=sqrt(q)
        t=acos(r/q/sq)
        s1=-2.*sq*cos(t/3.)-a1/3.
        s2=-2.*sq*cos((t+2.*pi)/3.)-a1/3.
        s3=-2.*sq*cos((t+4.*pi)/3.)-a1/3.
        if (s1.lt.0.) s1=1.d10
        if (s2.lt.0.) s2=1.d10
        if (s3.lt.0.) s3=1.d10
        ell=min(s1,s2,s3)
        if (ell.eq.1.d10) ell=-.1
     endif
  endif
  if (del.gt.0..and.ell.gt.0.) then
     corr=-.364/del*exp(-6.5*(l1-l2)/del-2.8*(l2-l3)/del)
  else
     corr=0.
  endif
  ell=ell+corr
  return

end function ell


subroutine ord(a,b,c)

! orders a,b,c in decreasing order a>b>c
  use numbers
  implicit none
  real(kind=doubleR),intent(inout) :: a,b,c
  real(kind=doubleR)::a1,b1,c1
  
  a1=max(a,b,c)
  c1=min(a,b,c)
  b1=a+b+c-a1-c1
  a=a1
  b=b1
  c=c1
  return

end subroutine ord


subroutine compute_velocities(ismooth)

! Computes the collapse times for one smoothing radius and position
! by collecting information from all grids
! and then computing the collapse times 

  use pinocchio_module
  use fmax_grids
  use fmax_smoothing_radii
  use final_products
  use fmax_main_memory

  implicit none
  integer(kind=singleI),intent(in)::ismooth

! local variables
  integer(kind=singleI)::local_z,global_x,global_y,global_z

! loop on all particles
  do local_z=0,Lgrid_local(3)-1
     global_z=local_z+Lgrid_z_start
     do global_y=0,Lgrid_local(2)-1
        do global_x=0,Lgrid_local(1)-1

#ifndef DONTSMOOTH_VEL
           if (Rmax(global_x,global_y,local_z) == ismooth) then
              vmax(:,global_x,global_y,local_z) = first_derivatives(:,global_x,global_y,local_z)
           endif
#else
           vmax(:,global_x,global_y,local_z) = first_derivatives(:,global_x,global_y,local_z)
#endif

        enddo
     enddo

! end of loop
  enddo

!  call message ('compute_velocities done')

! bye!
  return
end subroutine compute_velocities


subroutine write_out()

! output of the Fmax, vmax and Rmax matrices

  use pinocchio_module
  use fmax_smoothing_radii
  use fmax_grids
  use final_products
  use fmax_main_memory
  implicit none

  if (WriteFmax) call write_fmax()
  if (WriteVmax) call write_vmax()
  if (WriteRmax) call write_rmax()

  return
end subroutine write_out



! ********************************************************************* !

! The following subroutines are used to compute and write displacements
! in the format of a Gadget snapshot

subroutine compute_displacements()

  use pinocchio_module
  use fmax_smoothing_radii
  use fmax_main_memory, only: first_derivatives
  implicit none
  character(len=24),external :: fdate
  real(kind=doubleR)::Rsmooth

#ifdef TZA
  Rsmooth=3.0/Hubble100
#else
  Rsmooth=0.0
#endif

  if (Rsmooth>0.) then
     write(info,'("Displacements will be generated with Gaussian smoothing at a scale of ",f12.5," Mpc")')Rsmooth
     call message(info)
  else
     write(info,'("Displacements will be generated with no smoothing")')
     call message(info)
  endif

! present choices: no smoothing, redshift takes as the first one in the list

  write(info,'("[",a24,"] writing displacements at redshift ",f8.5,"...")') fdate(),zout(1)
  call message(info)

  call assign_fftw_variables(1)

  call allocate_VEL_for_displacements()

  write(info,"('Total allocated memory:',f8.3,' Gbyte; max allocated memory: ',f8.3,' Gbyte')") &
       total_allocated/Gbyte,max_allocated/Gbyte
  call message (trim(info))

  call allocate_first_derivatives()

  write(info,'("[",a24,"] computing displacements...")') fdate()
  call message(info)

  call compute_first_derivatives(Rsmooth, 1)

  call store_velocities(0,zout(1))

  call deallocate_first_derivatives()

  cputmp=MPI_WTIME()
  call write_snapshot(zout(1))
  cputime(5)=cputime(5)+MPI_WTIME()-cputmp

  call deallocate_VEL_for_displacements()

  write(info,"('Total allocated memory:',f8.3,' Gbyte; max allocated memory: ',f8.3,' Gbyte')") &
       total_allocated/Gbyte,max_allocated/Gbyte
  call message (trim(info))

  return
end subroutine compute_displacements



subroutine store_velocities(flag,z)

! it adds velocities to a VEL vector taking into account
! the growing mode for the various LPT terms

  use pinocchio_module
  use fmax_main_memory
  implicit none

  integer(kind=singleI),intent(in) :: flag
  real(kind=doubleR),intent(in) :: z

  real(kind=doubleR),external :: grow
  real(kind=doubleR) :: fac


  fac=grow(z)
  write(info,'("Growing mode for displacements at z=",f10.6,": ",f10.6,f10.6)')z,fac,1./fac
  call message(info)

  if (flag==0) then

     VEL_for_displacements=fac*first_derivatives

  endif

  return
end subroutine store_velocities


function pbc(x,L)

  use numbers
  real(kind=doubleR),intent(in) :: x,L
  real(kind=doubleR) :: pbc

  if (x<0) then 
     pbc = x+L
  else if (x>=L) then
     pbc = x-L
  else
     pbc = x
  endif

  return
end function pbc


subroutine write_cputimes()

  use pinocchio_module
  implicit none

  if (proc_id==0) then
     
     write(*,*)
     write(*,'("Table of CPU times")')
     write(*,'("Total:          ",g14.6)')cputime(0)
     write(*,'("Density in FS:  ",g14.6," (",f5.2,"%)")')cputime(4),100.*cputime(4)/cputime(0)
     write(*,'("FFTs:           ",g14.6," (",f5.2,"%)")')cputime(1),100.*cputime(1)/cputime(0)
     write(*,'("Collapse times: ",g14.6," (",f5.2,"%)")')cputime(2),100.*cputime(2)/cputime(0)
     write(*,'("Velocities:     ",g14.6," (",f5.2,"%)")')cputime(3),100.*cputime(3)/cputime(0)
     write(*,'("Total fmax:     ",g14.6," (",f5.2,"%)")')cputime(9),100.*cputime(9)/cputime(0)
     write(*,'("Redistribution: ",g14.6," (",f5.2,"%)")')cputime(6),100.*cputime(6)/cputime(0)
     write(*,'("Sorting:        ",g14.6," (",f5.2,"%)")')cputime(7),100.*cputime(7)/cputime(0)
     write(*,'("Fragmentation:  ",g14.6," (",f5.2,"%)")')cputime(8),100.*cputime(8)/cputime(0)
     write(*,'("Total fragment: ",g14.6," (",f5.2,"%)")')cputime(10),100.*cputime(10)/cputime(0)
     write(*,'("Total I/O:      ",g14.6," (",f5.2,"%)")')cputime(5),100.*cputime(5)/cputime(0)
     write(*,*)

  endif

  return
end subroutine write_cputimes
