!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

subroutine fmax()


  use pinocchio_module
  use fmax_smoothing_radii
  use fmax_grids
  implicit none

! local variables
  integer(kind=singleI):: ismooth, igrid

  character(len=24):: fdate
  external fdate

  cputime(9)=MPI_WTIME()

  if (ngrid>1) then

     call abort_code("Multiple grids are not implemented yet")

! here we should
!
!      1) initialize fftw for the large-scale grid (the same as the small-scale one?)
!      2) compute first derivatives of the large grid
!      3) compute second derivatives of the large grid
!      4) distribute to all processors and store the few used grid points
!      5) re-initialize fftw for the small-scale grid

  endif


  write(info,'("[",a24,"] First part: computation of collapse times")') fdate()
  call message(info)

! ******************************
! routines to initialize the run
! ******************************
  call initialization()

  write(info,'("[",a24,"] initialization done")') fdate()
  call message(info)

! if required, it loads Fmax and vmax from files and exits the fmax part
!  if (StartFromSavedFmax) then
!     call read_fmax()
!     call read_vmax()
!     call read_variances()
!     call deallocate_some_large_matrices()
!     return
!  endif

! if required, it writes out the density field
  if (WriteDensity) then
     write(info,'("[",a24,"] computing density...")') fdate()
     call message(info)

     call allocate_density()

     do igrid = 1, ngrid
        call compute_density(nsmooth,igrid)
     enddo

     cputmp=MPI_WTIME()
     write(info,'("[",a24,"] writing density...")') fdate()
     call write_density()
     cputime(5)=cputime(5)+MPI_WTIME()-cputmp

     call deallocate_density()
  endif


! if required, it generates displacements and writes them in a gadget-like format
  if (WriteDisplacements) then

     call compute_displacements()

  endif


! ****************************
! derivatives of the large box
! ****************************


! the following computations are all done for grid 1
  call assign_fftw_variables(1)

! ************************
! CYCLE ON SMOOTHING RADII
! ************************
  do ismooth = 1, nsmooth

     call message ("")
     write(info,'("[",a24,"] Starting smoothing radius ",i5," of ",i5," (R=",f9.5,", sigma=",f8.5,")")') fdate(),ismooth,nsmooth &
          ,Smoothing_Radii(ismooth),sqrt(Smoothing_Variances(ismooth))
     call message (trim(info))


! ***********************************************
! PART 1: 
! Compute the second derivatives of the potential
! ***********************************************

     call allocate_second_derivatives()

     cputmp = MPI_WTIME()            ! timing

     write(info,'(''['',a24,''] Computing second derivatives'')') fdate()
     call message (trim(info))
     call compute_second_derivatives(Smoothing_Radii(ismooth), 1)

     cputmp = MPI_WTIME()-cputmp     ! timing
     write(info,'(''['',a24,''] Done second derivatives, cpu time = '',g14.6)') fdate(), cputmp
     call message (trim(info))


! **************************
! PART 2:
! Compute the collapse times
! **************************

     cputmp = MPI_WTIME()

     write(info,'(''['',a24,''] Computing collapse times'')') fdate()
     call message (trim(info))

     call compute_collapse_times(ismooth)

     cputmp = MPI_WTIME() - cputmp    ! timing
     write(info,'(''['',a24,''] Done computing collapse times, cputime = '',g14.6)') fdate(),cputmp
     call message (trim(info))
     cputime(2)=cputime(2)+cputmp

! deallocate second derivative matrix

     write(info,"('Total allocated memory:',f8.3,' Gbyte; max allocated memory: ',f8.3,' Gbyte')") &
          total_allocated/Gbyte,max_allocated/Gbyte
     call message (trim(info))

     call deallocate_second_derivatives()

! ***********************************************
! PART 3: 
! computes the first derivatives of the potential
! ***********************************************

#ifdef DONTSMOOTH_VEL
     if (ismooth==nsmooth) then
#endif

     cputmp = MPI_WTIME()              ! timing

     call allocate_first_derivatives()

     write(info,'(''['',a24,''] Computing first derivatives'')') fdate()
     call message (trim(info))

     call compute_first_derivatives(Smoothing_Radii(ismooth), 1)

     cputmp = MPI_WTIME() - cputmp     ! timing
     write(info,'(''['',a24,''] Done computing first derivatives, cptime = '',g14.6)') fdate(),cputmp
     call message (trim(info))


! ********************************************
! PART 4:
! it stores velocities for collapsed particles
! ********************************************
     cputmp = MPI_WTIME()                 ! timing

! for a single grid calls the routine only once
     write(info,'(''['',a24,''] Computing velocities'')') fdate()
     call message (trim(info))

     call compute_velocities(ismooth,1)

     write(info,"('Total allocated memory:',f8.3,' Gbyte; max allocated memory: ',f8.3,' Gbyte')") &
          total_allocated/Gbyte,max_allocated/Gbyte
     call message (trim(info))

     call deallocate_first_derivatives()

     cputmp = MPI_WTIME() - cputmp         ! timing
     write(info,'(''['',a24,''] Done computing velocities, cputime = '',g14.6)') fdate(),cputmp
     call message (trim(info))
     cputime(3)=cputime(3)+cputmp

#ifdef DONTSMOOTH_VEL
     endif        
#endif


! *******************************
! END OF CYCLE ON SMOOTHING RADII
! *******************************
     call message ('Smoothing radius completed')
     call MPI_BARRIER(MPI_COMM_WORLD,ierr)
     
     write(info,'("expected and computed rms of the smoothed density: ",2(1x,f12.6))') &
          sqrt(Smoothing_Variances(ismooth)),sqrt(True_Variances(ismooth))
     call message(info)

  enddo


! ************************************************
! FINAL PART:
! Output density, Fmax and Vel* fields if required
! ************************************************

! Writes the results in files in the Data/ directory
  cputmp=MPI_WTIME()
  call write_out()
  cputime(5)=cputime(5)+MPI_WTIME()-cputmp

! Deallocates everything except Fmax and vmax
  call deallocate_some_large_matrices()

! final cpu time
  cputime(9) = MPI_WTIME() - cputime(9)
  write(info,'(''['',a24,''] Finishing fmax, total cputime = '',g14.6)') fdate(),cputime(9)
  call message (trim(info))

  return

end subroutine fmax


function fdate ()
  
  implicit none
  character(len=24) :: fdate
  character(len=8)  :: date
  character(len=10) :: time
  character(len=5)  :: timezone
  integer                   :: values(8)
  
  call date_and_time (date, time, timezone, values)
  
  write (fdate, &
       &   "(I2.2,'-',I2.2,'-',I4.4,' ',I2.2,':',I2.2,'.',I2.2,'     ')") &
       &   values(2), values(3), values(1), values(5), values(6), &
       &   values(7)
  
  return
end function fdate
                                                                                                                                     
 
